#!/usr/bin/env python

# stdlib imports
import os.path
import pathlib
from tempfile import mkdtemp, mkstemp
from unittest import mock

# third party imports
import pandas as pd

# local imports
from strec.utils import (
    create_config,
    get_config,
    get_config_file_name,
    read_input_file,
    render_row,
)


def move_config():
    strec_config_file = get_config_file_name()
    tempfile = None
    if os.path.isfile(strec_config_file):
        tempdir = mkdtemp()
        tempfile = os.path.join(tempdir, "strec.ini")
        os.rename(strec_config_file, tempfile)
    return tempfile


def restore_config(tempfile):
    strec_config_file = get_config_file_name()
    if tempfile is not None:
        tempdir, fname = os.path.split(tempfile)
        os.rename(tempfile, strec_config_file)
        os.rmdir(tempdir)


def get_config_dict(config):
    if not isinstance(config, dict):
        dictionary = {}
        for section in config.sections():
            dictionary[section] = {}
            for option in config.options(section):
                dictionary[section][option] = config.get(section, option)
    else:
        dictionary = config.copy()

    return dictionary


def test_get_config():
    # this tests a config file that may or may not be present
    # on a users system.  So, if one is there let's move it aside
    # so that testing is always consistent between systems.
    try:
        tempdir = mkdtemp()
        tempfile = os.path.join(tempdir, "config.ini")
        with mock.patch("strec.utils.get_config_file_name") as mocked_getfile:
            mocked_getfile.return_value = pathlib.Path(tempfile)
            create_config(tempdir)
            config = get_config()
            dictionary = get_config_dict(config)
            assert sorted(list(dictionary.keys())) == ["CONSTANTS", "DATA"]
            cmp_options = [
                "maxradial_disthist",
                "ddepth_intra",
                "depth_rangecomp",
                "ddepth_interf",
                "maxradial_distcomp",
                "dstrike_interf",
                "minradial_distcomp",
                "default_szdip",
                "dlambda",
                "minno_comp",
                "step_distcomp",
                "minradial_disthist",
                "ddip_interf",
            ]
            assert sorted(cmp_options) == sorted(config["CONSTANTS"].keys())
    except Exception as e:
        raise (e)
    finally:
        if pathlib.Path(tempfile).is_file():
            os.remove(tempfile)


def test_read_input_file():
    df = pd.DataFrame({"lat": [1, 2, 3], "lon": [1, 2, 3], "depth": [1, 2, 3]})
    tmpfile = None
    try:
        h, tmpfile = mkstemp(suffix=".xlsx")
        os.close(h)
        df.to_excel(tmpfile)
        read_input_file(tmpfile, [None], None)
        df.to_csv(tmpfile)
        read_input_file(tmpfile, [None], None)
        f = open(tmpfile, "wb")
        barray = bytearray([123, 3, 255, 0, 100])
        f.write(barray)
        f.close()
        try:
            _ = read_input_file(tmpfile, [None], None)
        except ValueError:
            assert 1 == 1
    except Exception as e:
        raise (e)
    finally:
        if tmpfile is not None:
            os.remove(tmpfile)


def test_render_row():
    row = pd.Series({"lat": 32.123, "lon": -118.123, "depth": 30.0})
    for pformat in ["pretty", "json", "csv"]:
        render_row(row, pformat, [None])


if __name__ == "__main__":
    test_get_config()
    test_render_row()
    test_read_input_file()



#!/usr/bin/env python

# stdlib imports
import json
import pathlib
import sys

# third party imports
from shapely.geometry import Point, shape

# local imports
from strec.distance import calc_distances, get_distance_to_regime


def print_dict(dict):
    for key, value in dict.items():
        print(f"{key} = {value:.1f} km")


def test_calc_tectonic_distances():

    points = [
        {
            "name": "Big Island",
            "region": "HotSpot",
            "lat": 19.630525,
            "lon": -155.435738,
            "order": [
                "DistanceToHotSpot",
                "DistanceToStable",
                "DistanceToActive",
                "DistanceToSubduction",
            ],
        },
        {
            "name": "Valparaiso, Chile",
            "region": "Subduction",
            "lat": -33.045879,
            "lon": -71.624954,
            "order": [
                "DistanceToSubduction",
                "DistanceToActive",
                "DistanceToHotSpot",
                "DistanceToStable",
            ],
        },
        {
            "name": "Coast of Greenland",
            "region": "Stable",
            "lat": 68.626508,
            "lon": -29.261091,
            "order": [
                "DistanceToStable",
                "DistanceToActive",
                "DistanceToHotSpot",
                "DistanceToSubduction",
            ],
        },
        {
            "name": "Tierra del Fuego",
            "region": "Active",
            "lat": -54.816815,
            "lon": -65.554447,
            "order": [
                "DistanceToActive",
                "DistanceToStable",
                "DistanceToHotSpot",
                "DistanceToSubduction",
            ],
        },
    ]

    for pdict in points:
        print(
            (
                f"Checking point ({pdict['lat']}, {pdict['lon']}) "
                f"{pdict['name']}, should be {pdict['region']}"
            )
        )
        distances = calc_distances(pdict["lat"], pdict["lon"])
        distances.pop("DistanceToOceanic", None)
        distances.pop("DistanceToContinental", None)
        sorted_distances = {
            k: v for k, v in sorted(distances.items(), key=lambda item: item[1])
        }
        cmp_order = list(sorted_distances.keys())
        print_dict(sorted_distances)
        print()
        assert pdict["order"] == cmp_order

    stable_points = [
        (40.964218, -95.890236),
        (35.958876, -83.608163),
        (52.633529, -105.613543),
        (51.771052, 18.998324),
        (8.105838, -6.461390),
        (-27.008106, 133.031058),
        (15.494769, 76.482348),
        (22.387245, 46.416856),
    ]
    for lat, lon in stable_points:
        distances = calc_distances(lat, lon)
        print(f"Checking stable point {lat},{lon}...")
        assert distances["DistanceToStable"] == 0

    active_points = [
        (35.701466, -120.115478),
        (21.300028, -77.624746),
        (29.077189, 26.729212),
        (36.175032, 55.557337),
        (-58.0, 149.7),
    ]
    for lat, lon in active_points:
        distances = calc_distances(lat, lon)
        assert distances["DistanceToActive"] == 0

    subduction_points = [
        (17.44, -98.49),
        (-15.52, -70.49),
        (34.39, -5.68),
        (35.70, 23.82),
        (-6.07, 126.74),
        (-39.50, 176.18),
        (45.93, 146.59),
        (51.81, 175.36),
    ]
    for lat, lon in subduction_points:
        distances = calc_distances(lat, lon)
        print(f"Checking subduction point {lat},{lon}...")
        assert distances["DistanceToSubduction"] == 0

    volcanic_points = [
        (19.723, -155.681),
        (45.73, -129.93),
        (-0.54, -90.03),
        (64.714, -18.394),
        (7.23, 38.37),
        (-63.76, -57.37),
        (-49.58, 69.12),
        (-37.64, 142.46),
    ]
    for lat, lon in volcanic_points:
        distances = calc_distances(lat, lon)
        assert distances["DistanceToHotSpot"] == 0


def test_calc_oceanic_distances():
    points = [
        {
            "name": "Madagascar",
            "region": "Continental",
            "lat": -13.356,
            "lon": 48.272,
            "order": [
                "DistanceToContinental",
                "DistanceToOceanic",
            ],
        },
        {
            "name": "United States",
            "region": "Continental",
            "lat": 43.41,
            "lon": -102.19,
            "order": [
                "DistanceToContinental",
                "DistanceToOceanic",
            ],
        },
        {
            "name": "Iceland",
            "region": "Oceanic",
            "lat": 64.878,
            "lon": -24.051,
            "order": [
                "DistanceToOceanic",
                "DistanceToContinental",
            ],
        },
        {
            "name": "South Pacific",
            "region": "Oceanic",
            "lat": -19.98,
            "lon": -105.38,
            "order": [
                "DistanceToOceanic",
                "DistanceToContinental",
            ],
        },
    ]
    for pdict in points:
        print(
            (
                f"Checking point ({pdict['lat']}, {pdict['lon']}) "
                f"{pdict['name']}, should be {pdict['region']}"
            )
        )
        distances = calc_distances(pdict["lat"], pdict["lon"])
        distances.pop("DistanceToActive", None)
        distances.pop("DistanceToStable", None)
        distances.pop("DistanceToSubduction", None)
        distances.pop("DistanceToHotSpot", None)
        sorted_distances = {
            k: v for k, v in sorted(distances.items(), key=lambda item: item[1])
        }
        cmp_order = list(sorted_distances.keys())
        print_dict(sorted_distances)
        print()
        assert pdict["order"] == cmp_order

    land_points = [
        (38.397633, -103.922577),
        (19.627053, -100.231171),
        (-5.457441, -61.559299),
        (21.110111, 12.444609),
        (-74.216469, 22.464139),
        (-22.608953, 124.944608),
        (-12.010521, 124.976613),
        (36.212620, 97.557150),
        (80.653060, 95.339942),
        (51.147136, 19.052844),
    ]
    for lat, lon in land_points:
        distances = calc_distances(lat, lon)
        assert distances["DistanceToContinental"] == 0.0

    ocean_points = [
        (46.393396, -138.365785),
        (44.107786, -125.987451),
        (-45.895658, -114.550334),
        (-20.972943, -74.749524),
        (-53.472616, 3.515289),
        (0, 0),
        (44.068439, -36.067652),
        (81.286241, -0.643259),
        (81.612268, 127.956193),
        (-43.702279, 155.658387),
    ]
    for lat, lon in ocean_points:
        distances = calc_distances(lat, lon)
        assert distances["DistanceToOceanic"] == 0.0


def test_get_regime_distance():
    # these are in lon,lat order
    hotspots = [
        (-30, 38.6),
        (-16.71, 32.83),
        (-16.71, 28.99),
        (-24.42, 15.78),
        (-29.08, -20.04),
        (-80.49, -26.01),
        (-89.46, -0.05),
        (-170.22, -14.30),
        (-111.05, 44.45),
    ]

    stables = [
        (-32.0, -63.298),
        (-62.13, -23.50),
        (-109.70, 43.80),
        (-118.57, 89.14),
        (-55.92, -89.75),
        (-60.0, -15.0),
    ]

    actives = [
        (-14.55, -8.80),
        (-17.648, -11.407),
        (-62.870, -23.630),
        (-110.91, 42.91),
    ]

    subductions = [
        (-74.316, -36.624),
        (-30.006, -56.368),
        (121.053, -11.288),
        (120.351, 20.814),
        (-179.960, 51.273),
        (179.955, 51.150),
    ]

    for lon, lat in hotspots:
        print(f"Checking ({lat},{lon}) is hotspot...")
        distance = get_distance_to_regime(lat, lon, "volcanic")
        assert distance == 0.0
    for lon, lat in stables:
        print(f"Checking ({lat},{lon}) is stable...")
        distance = get_distance_to_regime(lat, lon, "stable")
        assert distance == 0.0
    for lon, lat in actives:
        print(f"Checking ({lat},{lon}) is active...")
        distance = get_distance_to_regime(lat, lon, "active")
        assert distance == 0.0
    for lon, lat in subductions:
        print(f"Checking ({lat},{lon}) is subduction...")
        distance = get_distance_to_regime(lat, lon, "subduction")
        assert distance == 0.0


def test_distances_along_lat():
    tlat = -21.326
    points = [
        (-65.531, tlat, "DistanceToSubduction"),
        (-63.173, tlat, "DistanceToSubduction"),
        (-62.904, tlat, "DistanceToSubduction"),
        (-62.652, tlat, "DistanceToActive"),
        (-62.242, tlat, "DistanceToActive"),
        (-61.960, tlat, "DistanceToActive"),
        (-61.837, tlat, "DistanceToStable"),
        (-61.656, tlat, "DistanceToStable"),
        (-61.445, tlat, "DistanceToStable"),
        (-61.122, tlat, "DistanceToStable"),
        (-60.378, tlat, "DistanceToStable"),
        (-59.822, tlat, "DistanceToStable"),
        (-56.070, tlat, "DistanceToStable"),
        (-58.57, 13.66, "DistanceToSubduction"),
        (-57.16, 13.66, "DistanceToActive"),
        (-55.93, 13.66, "DistanceToStable"),
    ]
    for lon, lat, rtype in points:
        distances = calc_distances(lat, lon)
        distances.pop("DistanceToContinental")
        distances.pop("DistanceToOceanic")
        ctype = min(distances, key=distances.get)
        print(min(distances.values()))
        if rtype != ctype:
            rdist = distances[rtype]
            cdist = distances[ctype]
            print(
                f"Point ({lat},{lon}) should be {rtype} ({rdist:.2f}), came back as {ctype} ({cdist:.2f})."
            )


if __name__ == "__main__":
    test_calc_tectonic_distances()
    test_distances_along_lat()
    test_get_regime_distance()
    test_calc_tectonic_distances()
    test_calc_oceanic_distances()



#!/usr/bin/env python

import numpy as np
from strec.kagan import get_kagan_angle


def test_kagan_consistency():
    # from https://earthquake.usgs.gov/earthquakes/eventpage/us2000ar20#moment-tensor
    mww = (108, 47, -98)
    mwc = (111, 46, -94)
    mwb = (311, 47, -85)

    mww_mwc = get_kagan_angle(mww[0], mww[1], mww[2], mwc[0], mwc[1], mwc[2])
    mww_mwb = get_kagan_angle(mww[0], mww[1], mww[2], mwb[0], mwb[1], mwb[2])
    mwc_mwb = get_kagan_angle(mwc[0], mwc[1], mwc[2], mwb[0], mwb[1], mwb[2])

    np.testing.assert_almost_equal(mww_mwc, 3.07, decimal=1)
    np.testing.assert_almost_equal(mww_mwb, 14.43, decimal=1)
    np.testing.assert_almost_equal(mwc_mwb, 13.95, decimal=1)


def test_kagan_slab():
    events = {'Sumatra': (336, 7, 114),
              'Tohoku': (193, 9, 78),
              'Maule': (178, 77, 86),
              'Illapel': (353, 19, 83)}
    # TODO - get slab values for these events
    #tuple is depth,dip,strike
    slabs = {'Sumatra': (38.66, 20.45, 307.08),
             'Tohoku': (30.23, 14.89, 190.22),
             'Maule': (29.33, 19.64, 378.94),
             'Illapel': (29.09, 18.60, 364.10)}
    for eqname, eqparams in events.items():
        event_strike, event_dip, event_rake = eqparams
        slabparams = slabs[eqname]
        slab_depth, slab_dip, slab_strike = slabparams
        kagan = get_kagan_angle(event_strike, event_dip,
                                event_rake, slab_strike, slab_dip, 90)
        print('%s event: Kagan Angle %.2f' % (eqname, kagan))


if __name__ == '__main__':
    test_kagan_consistency()
    test_kagan_slab()



# stdlib imports
import gzip
import io
import logging
import os.path
import tempfile
from datetime import datetime
from urllib import parse, request

# third party libraries
import numpy as np
import pandas as pd

TIMEOUT = 30
HIST_GCMT_URL = (
    "http://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/jan76_dec17.ndk.gz"
)
MONTHLY_GCMT_URL = (
    "http://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/NEW_MONTHLY/"
)


def fetch_gcmt():
    """Fetch all GCMT data from gcmt web site, return into pandas DataFrame.

    Returns:
        pandas DataFrame containing columns:
            - time (YYYY-MM-DD HH:MM:SS for CSV)
            - lat (decimal degrees)
            - lon (decimal degrees)
            - depth (km)
            - mag Magnitude
            - mrr Mrr moment tensor component
            - mtt Mtt moment tensor component
            - mpp Mpp moment tensor component
            - mrt Mrt moment tensor component
            - mrp Mrp moment tensor component
            - mtp Mtp moment tensor component
    """
    t1 = str(datetime.utcnow())
    logging.info("%s - Fetching historical GCMT data..." % t1)
    histfile = get_historical_gcmt()
    t2 = str(datetime.utcnow())
    logging.debug("%s - Converting to dataframe..." % t2)
    dataframe = ndk_to_dataframe(histfile)
    t3 = str(datetime.utcnow())
    logging.info("%s - Fetching monthly data..." % t3)
    start_year = 2018
    end_year = datetime.utcnow().year
    end_month = datetime.utcnow().month
    for year in range(start_year, end_year + 1):
        for month in range(1, 13):
            if year == end_year and month > end_month:
                continue
            logging.debug("Fetching GCMT data for %i month %i..." % (year, month))
            monthfile = get_monthly_gcmt(year, month)
            if monthfile is None:
                logging.debug("No NDK file for %i month %i" % (year, month))
                continue
            month_frame = ndk_to_dataframe(monthfile)
            dataframe = pd.concat([dataframe, month_frame], axis=0)
    return dataframe


def get_historical_gcmt():
    """Retrieve the Jan 1976 - Dec 2010 GCMT catalog.

    NDK format explained:

    http://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/allorder.ndk_explained

    Returns:
        io.StringIO:
            StringIO object containing the NDK file for the 1976-2010 period.
    """
    zipfile = None
    try:
        fh = request.urlopen(HIST_GCMT_URL)
        zip_bytes = fh.read()
        fh.close()
        handle, zipfile = tempfile.mkstemp()
        os.close(handle)
        f = open(zipfile, "wb")
        f.write(zip_bytes)
        f.close()
        gz = gzip.GzipFile(zipfile, mode="rb")
        unzip_bytes = gz.read().decode("utf-8")
        string_unzip = io.StringIO(unzip_bytes)
        return string_unzip
    except Exception as msg:
        raise Exception(msg)
    finally:
        if zipfile is not None:
            os.remove(zipfile)


def get_monthly_gcmt(year, month):
    """Download one month's worth of GCMT data into a StringIO object.

    Args:
        year (int):
            Integer year.
        month (int):
            Integer month (1-12).
    Returns:
        io.StringIO:
            StringIO object containing NDK file for given month/year.
    """
    strmonth = [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ][month - 1]
    stryear = str(year)
    url = parse.urljoin(MONTHLY_GCMT_URL, f"{stryear}/{strmonth}{stryear[2:]}.ndk")

    try:
        fh = request.urlopen(url, timeout=TIMEOUT)
        bytes = fh.read().decode("utf-8")
        fh.close()
        stringfile = io.StringIO(bytes)
        return stringfile
    except Exception:
        return None


def ndk_to_dataframe(ndkfile):
    """Turn an ndk file-like object or filename into a pandas dataframe.

    Args:
        ndkfile (str or file-like object):
            String file name or file-like object containing NDK-formatted earthquake
            data.
    Returns:
        pandas DataFrame containing columns:
            - time (YYYY-MM-DD HH:MM:SS for CSV)
            - lat (decimal degrees)
            - lon (decimal degrees)
            - depth (km)
            - mag Magnitude
            - mrr Mrr moment tensor component
            - mtt Mtt moment tensor component
            - mpp Mpp moment tensor component
            - mrt Mrt moment tensor component
            - mrp Mrp moment tensor component
            - mtp Mtp moment tensor component
    """
    if not hasattr(ndkfile, "read"):
        ndkfile = open(ndkfile, "r")

    lc = 0
    df = pd.DataFrame(
        columns=[
            "time",
            "lat",
            "lon",
            "depth",
            "mag",
            "mrr",
            "mtt",
            "mpp",
            "mrt",
            "mrp",
            "mtp",
        ]
    )
    rows = []
    tdict = {}
    for line in ndkfile.readlines():
        if (lc + 1) % 5 == 1:
            _parse_line1(line, tdict)
            lc += 1
            continue
        if (lc + 1) % 5 == 2:
            lc += 1
            continue
        if (lc + 1) % 5 == 3:
            lc += 1
            continue
        if (lc + 1) % 5 == 4:
            _parse_line4(line, tdict)
            lc += 1
            continue
        if (lc + 1) % 5 == 0:
            _parse_line5(line, tdict)
            lc += 1
            tdict.pop("exponent")  # remove now extraneous field
            rows.append(tdict)
            tdict = {}
            continue

    ndkfile.close()
    df = pd.DataFrame(rows)
    return df


def _parse_line1(line, tdict):
    """Parse the first line of an NDK file."""
    dstr = line[5:26]
    year = int(dstr[0:4])
    month = int(dstr[5:7])
    day = int(dstr[8:10])
    hour = int(dstr[11:13])
    minute = int(dstr[14:16])
    fseconds = float(dstr[17:])
    seconds = int(fseconds)
    if seconds > 59:
        seconds = 59
    microseconds = int((fseconds - seconds) * 1e6)
    if microseconds > 999999:
        microseconds = 999999

    tdict["time"] = datetime(year, month, day, hour, minute, seconds, microseconds)

    tdict["lat"] = float(line[27:33])
    tdict["lon"] = float(line[34:41])
    tdict["depth"] = float(line[42:47])


def _parse_line4(line, tdict):
    """Parse the fourth line of an NDK file."""
    tdict["exponent"] = float(line[0:2])
    tdict["mrr"] = float(line[2:9]) * np.power(10.0, tdict["exponent"])
    tdict["mtt"] = float(line[15:22]) * np.power(10.0, tdict["exponent"])
    tdict["mpp"] = float(line[28:35]) * np.power(10.0, tdict["exponent"])
    tdict["mrt"] = float(line[41:48]) * np.power(10.0, tdict["exponent"])
    tdict["mrp"] = float(line[54:61]) * np.power(10.0, tdict["exponent"])
    tdict["mtp"] = float(line[67:74]) * np.power(10.0, tdict["exponent"])


def _parse_line5(line, tdict):
    """Parse the fifth line of an NDK file."""
    scalar_moment = float(line[49:56].strip()) * np.power(10.0, tdict["exponent"])
    tdict["mag"] = ((2.0 / 3.0) * np.log10(scalar_moment)) - 10.7



#!/usr/bin/env python

from strec.gmreg import Regionalizer


def test_regionalizer():
    regionalizer = Regionalizer.load()

    # these are in lon,lat order
    hotspots = [
        (-30, 38.6),
        (-16.71, 32.83),
        (-16.71, 28.99),
        (-24.42, 15.78),
        (-29.08, -20.04),
        (-80.49, -26.01),
        (-89.46, -0.05),
        (-170.22, -14.30),
        (-111.05, 44.45),
    ]

    stables = [
        (-32.0, -63.298),
        (-62.13, -23.50),
        (-109.70, 43.80),
        (-118.57, 89.14),
        (-55.92, -89.75),
        (-60.0, -15.0),
    ]

    actives = [
        (-14.55, -8.80),
        (-17.648, -11.407),
        (-62.870, -23.630),
        (-110.91, 42.91),
    ]

    subductions = [
        (-74.316, -36.624),
        (-30.006, -56.368),
        (121.053, -11.288),
        (120.351, 20.814),
        (-179.960, 51.273),
        (179.955, 51.150),
    ]
    for lon, lat in hotspots:
        print(f"Checking ({lat},{lon}) is hotspot...")
        results = regionalizer.getRegions(lat, lon)
        assert results["TectonicRegion"] == "HotSpot"
    for lon, lat in stables:
        print(f"Checking ({lat},{lon}) is stable...")
        results = regionalizer.getRegions(lat, lon)
        assert results["TectonicRegion"] == "Stable"
    for lon, lat in actives:
        print(f"Checking ({lat},{lon}) is active...")
        results = regionalizer.getRegions(lat, lon)
        assert results["TectonicRegion"] == "Active"
    for lon, lat in subductions:
        print(f"Checking ({lat},{lon}) is subduction...")
        results = regionalizer.getRegions(lat, lon)
        assert results["TectonicRegion"] == "Subduction"


if __name__ == "__main__":
    test_regionalizer()



#!/usr/bin/env python

# stdlib imports
import pathlib

# third party imports
import numpy as np
import pandas as pd
import pytest

# local imports
from strec.calc import (
    check_moment_row,
    get_input_dataframe,
    get_moment_columns,
    select_regions,
)

ATOL = 0.2
RTOL = 0.9


def compare_tensors(tensor1, tensor2):
    for key, value in tensor1.items():
        cmp_value = tensor2[key]
        if isinstance(value, str):
            assert value == cmp_value
        elif isinstance(value, dict):
            for subkey, subvalue in value.items():
                np.testing.assert_allclose(
                    subvalue, cmp_value[subkey], rtol=RTOL, atol=ATOL
                )

        else:
            np.testing.assert_allclose(value, cmp_value, rtol=RTOL, atol=ATOL)


def test_get_moment_columns():
    row = pd.Series(
        {
            "Mrr": 147400000000000,
            "Mtt": -534500000000000,
            "Mpp": 387100000000000,
            "Mrp": 282800000000000,
            "Mrt": -65680000000000,
            "Mtp": -249700000000000,
        }
    )
    tensor_params = get_moment_columns(row)
    cmp_params = {
        "mrr": 147400000000000.0,
        "mtt": -534500000000000.0,
        "mpp": 387100000000000.0,
        "mrt": -65680000000000.0,
        "mrp": 282800000000000.0,
        "mtp": -249700000000000.0,
        "source": "unknown",
        "type": "unknown",
        "NP1": {
            "strike": 296.30077024634284,
            "dip": 68.11685306165077,
            "rake": 157.15208404514294,
        },
        "NP2": {
            "strike": 35.225709774290436,
            "dip": 68.88083955258415,
            "rake": 23.55014283538641,
        },
        "T": {
            "azimuth": 255.93919800048124,
            "value": 626337818224371.0,
            "plunge": 31.22002840266667,
        },
        "N": {
            "azimuth": 74.80537317180232,
            "value": -28481000172499.133,
            "plunge": 58.77499824283217,
        },
        "P": {
            "azimuth": 165.63454565853377,
            "value": -597856818051871.8,
            "plunge": 0.5026289265483992,
        },
    }
    compare_tensors(tensor_params, cmp_params)
    row = pd.Series(
        {
            "Strike": 1,
            "Dip": 1,
            "Rake": 1,
        }
    )
    tensor_params = get_moment_columns(row)
    cmp_params = {
        "mrr": 766786542744639.8,
        "mtt": -766903310262146.4,
        "mpp": 116767517506.37914,
        "mrt": -1.2587334961693614e18,
        "mrp": 10036448003488.078,
        "mtp": -2.196792959061808e16,
        "source": "unknown",
        "type": "unknown",
        "NP1": {
            "strike": 0.9999999997718305,
            "dip": 0.9999999997715706,
            "rake": 0.9999999997715463,
        },
        "NP2": {
            "strike": 270.00015227392294,
            "dip": 89.98254847933853,
            "rake": 90.99984772584894,
        },
        "T": {
            "azimuth": 181.00015232031038,
            "value": 1.2589254117941714e18,
            "plunge": 45.00872576013233,
        },
        "N": {
            "azimuth": 89.99984770289579,
            "value": -0.17311440291958194,
            "plunge": 0.999847679693138,
        },
        "P": {
            "azimuth": 359.00076099885496,
            "value": -1.2589254117941714e18,
            "plunge": 44.97382803145975,
        },
    }
    compare_tensors(tensor_params, cmp_params)


def test_check_moment_row():
    row = pd.Series(
        {
            "Mrr": 1,
            "Mtt": 1,
            "Mpp": 1,
            "Mrp": 1,
            "Mrt": 1,
            "Mtp": 1,
        }
    )
    assert check_moment_row(row)
    row = pd.Series(
        {
            "Strike": 1,
            "Dip": 1,
            "Rake": 1,
        }
    )
    assert check_moment_row(row)
    row = pd.Series(
        {
            "Dip": 1,
            "Rake": 1,
        }
    )
    assert not check_moment_row(row)


def test_get_input_dataframe():
    simple_catalog_file = pathlib.Path(__file__).parent / "data" / "simple_catalog.xlsx"
    eqinfo = None
    event_id = None
    hypo_columns = [None, None, None]
    id_column = None
    simple_dataframe, _, _, _, _ = get_input_dataframe(
        simple_catalog_file, eqinfo, event_id, hypo_columns, id_column
    )
    cmp_simple = pd.DataFrame(
        {
            "Latitude": [-5.074, -1.008],
            "Longitude": [103.083, 98.642],
            "Depth": [50.5, 17.6],
        }
    )
    assert simple_dataframe.equals(cmp_simple)

    comcat_catalog_file = (
        pathlib.Path(__file__).parent / "data" / "comcatid_catalog.xlsx"
    )
    eqinfo = None
    event_id = None
    hypo_columns = [None, None, None]
    id_column = None
    comcat_dataframe, idcol, _, _, _ = get_input_dataframe(
        comcat_catalog_file, eqinfo, event_id, hypo_columns, id_column
    )
    assert idcol == "EventID"

    id_catalog_file = pathlib.Path(__file__).parent / "data" / "id_catalog.xlsx"
    eqinfo = None
    event_id = None
    hypo_columns = [None, None, None]
    id_column = "id"
    comcat_dataframe, idcol, _, _, _ = get_input_dataframe(
        id_catalog_file, eqinfo, event_id, hypo_columns, id_column
    )
    assert idcol == "id"

    hypo_catalog_file = pathlib.Path(__file__).parent / "data" / "hypocols_catalog.xlsx"
    eqinfo = None
    event_id = None
    hypo_columns = ["EventLatitude", "EventLongitude", "EventDepth"]
    id_column = None
    hypo_dataframe, idcol, latcol, loncol, depcol = get_input_dataframe(
        hypo_catalog_file, eqinfo, event_id, hypo_columns, id_column
    )
    assert [latcol, loncol, depcol] == hypo_columns

    moment_catalog_file = pathlib.Path(__file__).parent / "data" / "moment_catalog.xlsx"
    eqinfo = None
    event_id = None
    hypo_columns = [None, None, None]
    id_column = None
    moment_dataframe, _, _, _, _ = get_input_dataframe(
        moment_catalog_file, eqinfo, event_id, hypo_columns, id_column
    )
    assert "mrt" in moment_dataframe.columns

    focal_catalog_file = pathlib.Path(__file__).parent / "data" / "focal_catalog.xlsx"
    eqinfo = None
    event_id = None
    hypo_columns = [None, None, None]
    id_column = None
    focal_dataframe, _, _, _, _ = get_input_dataframe(
        focal_catalog_file, eqinfo, event_id, hypo_columns, id_column
    )
    assert "strike" in focal_dataframe.columns


def test_select_regions():
    simple_catalog_file = pathlib.Path(__file__).parent / "data" / "simple_catalog.xlsx"
    eqinfo = None
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = [None, None, None]
    id_column = None
    simple_dataframe = select_regions(
        simple_catalog_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    assert simple_dataframe["TectonicRegion"].unique().tolist() == ["Subduction"]
    assert (simple_dataframe["KaganAngle"] < 15).all()

    comcat_catalog_file = (
        pathlib.Path(__file__).parent / "data" / "comcatid_catalog.xlsx"
    )
    eqinfo = None
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = [None, None, None]
    id_column = None
    comcat_dataframe = select_regions(
        comcat_catalog_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    assert comcat_dataframe["FocalMechanism"].to_list() == ["RS", "SS"]

    id_catalog_file = pathlib.Path(__file__).parent / "data" / "id_catalog.xlsx"
    eqinfo = None
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = [None, None, None]
    id_column = "id"
    id_dataframe = select_regions(
        id_catalog_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    assert id_dataframe["FocalMechanism"].to_list() == ["RS", "SS"]

    hypo_catalog_file = pathlib.Path(__file__).parent / "data" / "hypocols_catalog.xlsx"
    eqinfo = None
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = ["EventLatitude", "EventLongitude", "EventDepth"]
    id_column = None
    hypo_dataframe = select_regions(
        hypo_catalog_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    assert hypo_dataframe["FocalMechanism"].to_list() == ["RS", "SS"]

    moment_catalog_file = pathlib.Path(__file__).parent / "data" / "moment_catalog.xlsx"
    eqinfo = None
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = [None, None, None]
    id_column = None
    moment_dataframe = select_regions(
        moment_catalog_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    assert moment_dataframe["FocalMechanism"].to_list() == ["RS", "SS"]

    focal_catalog_file = pathlib.Path(__file__).parent / "data" / "focal_catalog.xlsx"
    eqinfo = None
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = [None, None, None]
    id_column = None
    focal_dataframe = select_regions(
        focal_catalog_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    assert focal_dataframe["FocalMechanism"].to_list() == ["RS", "SS"]

    eqinfo = [37.7132, 141.5793, 41]
    event_id = None
    moment_info = [None, None, None, None]
    hypo_columns = [None, None, None]
    id_column = None
    eq_dataframe = select_regions(
        None,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        False,
    )
    eq_dataframe.iloc[0]["FocalMechanism"] == "RS"


if __name__ == "__main__":
    test_get_moment_columns()
    test_check_moment_row()
    test_get_input_dataframe()
    test_select_regions()



#!/usr/bin/env python

# stdlib imports
import pathlib
from collections import OrderedDict

# third party
import pandas as pd

# local imports
from strec.distance import calc_distances
from strec.utils import get_config


class Regionalizer(object):
    def __init__(self, datafolder):
        """Determine tectonic region information given epicenter and depth.

        Args:
            datafolder (str): Path to directory containing spatial data
            for tectonic regions.
        """
        self._datafolder = pathlib.Path(datafolder)
        config = get_config()
        grid_data_folder = pathlib.Path(config["DATA"]["folder"])
        self._tectonic_grid = grid_data_folder / "tectonic_global.tif"
        self._oceanic_grid = grid_data_folder / "oceanic_global.tif"

    @classmethod
    def load(cls):
        """Load regionalizer data from data in the repository.

        Returns:
            Regionalizer: Instance of Regionalizer class.
        """
        config = get_config()
        datadir = config["DATA"]["folder"]
        return cls(datadir)

    def getRegions(self, lat, lon):
        """Get information about the tectonic region of a given hypocenter.

        Args:
            lat (float): Earthquake hypocentral latitude.
            lon (float): Earthquake hypocentral longitude.
            depth (float): Earthquake hypocentral depth.
        Returns:
            Series: Pandas series object containing labels:
                - TectonicRegion: Subduction, Active, Stable, or HotSpot.
                - DistanceToStable: Distance in km to nearest stable region.
                - DistanceToActive: Distance in km to nearest active region.
                - DistanceToSubduction: Distance in km to nearest subduction
                                        region.
                - DistanceToHotSpot: Distance in km to nearest hotspot
                                      region.
                - Oceanic: Boolean indicating if epicenter is in the ocean.
                - DistanceToOceanic: Distance in km to nearest oceanic region.
                - DistanceToContinental: Distance in km to nearest continental
                                         region.
        """
        regions = OrderedDict()

        region_dict = calc_distances(lat, lon)

        if region_dict["DistanceToActive"] == 0:
            region_dict["TectonicRegion"] = "Active"
        elif region_dict["DistanceToStable"] == 0:
            region_dict["TectonicRegion"] = "Stable"
        elif region_dict["DistanceToSubduction"] == 0:
            region_dict["TectonicRegion"] = "Subduction"
        else:
            region_dict["TectonicRegion"] = "HotSpot"

        region_dict["Oceanic"] = False
        if region_dict["DistanceToOceanic"] == 0:
            region_dict["Oceanic"] = True

        regions = pd.Series(
            region_dict,
            index=[
                "TectonicRegion",
                "DistanceToStable",
                "DistanceToActive",
                "DistanceToSubduction",
                "DistanceToHotSpot",
                "Oceanic",
                "DistanceToOceanic",
                "DistanceToContinental",
            ],
        )

        return regions



#!/usr/bin/env python

# stdlib imports
import os.path
import pathlib
from datetime import datetime

# third party imports
from esi_utils_rupture.tensor import fill_tensor_from_components

# remove later
from strec.gcmt import fetch_gcmt, ndk_to_dataframe
from strec.subtype import get_focal_mechanism


def throw_away():
    dataframe = fetch_gcmt()
    mechs = []
    for idx, row in dataframe.iterrows():
        tensor = fill_tensor_from_components(
            row["mrr"], row["mtt"], row["mpp"], row["mrt"], row["mrp"], row["mtp"]
        )
        mech = get_focal_mechanism(tensor)
        mechs.append(mech)
    dataframe["focal_mechanism"] = mechs
    try:
        dataframe.to_excel("gcmt_mechanisms.xlsx")
    except Exception as e:
        raise e
    finally:
        os.remove("gcmt_mechanisms.xlsx")


def null_test_fetch_gcmt():
    dataframe = fetch_gcmt()
    first = dataframe.iloc[0]
    assert first["time"] == datetime(1976, 1, 1, 1, 29, 39, 600000)
    assert first["lat"] == -28.61
    print(first["mrr"])
    assert first["mrr"] == 7.68e26
    random = dataframe.iloc[40000]
    assert random["time"] == datetime(2012, 11, 23, 0, 40, 21, 199999)
    assert random["lat"] == 1.62
    print(random["mrr"])
    assert random["mrr"] == 8.01e22


def test_ndk_read():
    homedir = pathlib.Path(__file__).parent  # where is this script?
    ndkfile = homedir / ".." / "data" / "test.ndk"
    dataframe = ndk_to_dataframe(ndkfile)
    assert dataframe.iloc[0]["time"] == datetime(2017, 1, 1, 14, 12, 8, 99999)
    assert dataframe.iloc[0]["lat"] == 3.66
    assert dataframe.iloc[0]["mrr"] == 2.460000e24


if __name__ == "__main__":
    # throw_away()
    test_ndk_read()
    # test_fetch_gcmt()



#!/usr/bin/env python

# stdlib imports
import sqlite3

# third party imports
import numpy as np

# local imports
from esi_utils_rupture.tensor import fill_tensor_from_components


def getComposite(rows):
    """Calculate composite moment tensor.

    Args:
        rows (list):List of tuples containing (mrr,mtt,mpp,mrt,mrp,mtp)
    Returns:
        tuple: Tuple of (composite moment tensor dictionary
                (see fill_tensor_from_angles),
                (scalar) similarity index,
                Number of rows used to calculate composite)
    """
    components = np.array(rows)
    components[:, 4] *= -1
    components[:, 5] *= -1
    nrows, ncols = components.shape
    # get max value in each row and tile to shape nrowsxncols
    max_values = np.tile(np.reshape(components.max(axis=1), (nrows, 1)), ncols)
    comp_norm = components / max_values
    comp_av = np.mean(comp_norm, axis=0)
    comp_av_tile = np.tile(np.reshape(comp_av, (ncols, 1)), nrows).transpose()
    comp_sq = np.power((comp_norm - comp_av_tile), 2)
    comp_sq_mean_sq = np.sqrt(np.mean(comp_sq, axis=0))

    m33 = comp_av[0]
    m11 = comp_av[1]
    m22 = comp_av[2]
    m13 = comp_av[3]
    m23 = comp_av[4]
    m12 = comp_av[5]

    vm11 = comp_sq_mean_sq[1]
    vm22 = comp_sq_mean_sq[2]
    vm13 = comp_sq_mean_sq[3]
    vm23 = comp_sq_mean_sq[4]
    vm12 = comp_sq_mean_sq[5]
    varforbenius = vm11 * vm11 + vm12 * vm12 + vm13 * vm13 + vm22 * vm22 + vm23 * vm23
    forbenius = m11 * m11 + m12 * m12 + m13 * m13 + m22 * m22 + m23 * m23
    similarity = np.sqrt(varforbenius) / forbenius

    mrr = m33
    mtt = m11
    mpp = m22
    mrt = m13
    mrp = -m23
    mtp = -m12

    tensor = fill_tensor_from_components(mrr, mtt, mpp, mrt, mrp, mtp)
    return (tensor, similarity, nrows)


def getCompositeCMT(lat, lon, dbfile, box=0.1, nmin=3, maxbox=1.0, dbox=0.09):
    """Search a database for list of moment tensors, calculate composite moment tensor.
    Args:
        lat (float): Latitude (dd).
        lon (float): Longitude (dd).
        dbfile (str): Path to sqlite database file.
        box (float): half-width of latitude/longitude search box (dd)
        depthbox (float): half-width of depth search window (km)
        nmin (int): Minimum number of events to use to calculate composite moment tensor
        maxbox (float): Maximum size of search box (dd)
        dbox (float): Increment of search box (dd)
    Returns:
        tuple: Tuple of (composite moment tensor dictionary
                (see fill_tensor_from_angles),
                (scalar) similarity index,
                Number of rows used to calculate composite)
    """
    conn = sqlite3.connect(dbfile)
    cursor = conn.cursor()
    rows = []
    searchwidth = box
    while len(rows) < nmin and searchwidth < maxbox:
        qstr = (
            "SELECT mrr,mtt,mpp,mrt,mrp,mtp FROM earthquake WHERE lat >= %.4f AND "
            "lat <= %.4f AND lon >= %.4f AND lon <= %.4f"
        )
        query = qstr % (
            lat - searchwidth,
            lat + searchwidth,
            lon - searchwidth,
            lon + searchwidth,
        )
        cursor.execute(query)
        rows = cursor.fetchall()
        if len(rows) >= nmin:
            break
        searchwidth += dbox

    if len(rows) == 0:
        if len(rows) > 0:
            edict, similarity, nrows = getComposite(rows)
            return (edict, similarity, nrows)
        else:
            return (None, np.nan, 0)

    edict, similarity, nrows = getComposite(rows)

    return (edict, similarity, nrows)



#!/usr/bin/env python

# local imports
from strec.subduction import SubductionZone, norm_angle


def test_norm_angle():
    angle1 = norm_angle(375)
    assert angle1 == 15
    angle2 = norm_angle(-15)
    assert angle2 == 345
    angle3 = norm_angle(725)
    assert angle3 == 5


def test_rupture_plane():
    config = {
        "CONSTANTS": {
            "dstrike_interf": 30,
            "ddip_interf": 30,
            "dlambda": 60,
            "ddepth_interf": 20,
            "ddepth_intra": 10,
        }
    }
    # using slab information from tohoku epicenter
    # and MT info from duputel solution for tohoku
    slab_params = {
        "strike": 190.22474670410156,
        "dip": 14.894770622253418,
        "depth": 30.23206329345703,
    }

    tensor_params = {
        "NP1": {"strike": 193, "dip": 9, "rake": 78},
        "NP2": {"strike": 25, "dip": 81, "rake": 92},
        "P": {"azimuth": 113, "plunge": 36},
    }

    depth = 29.0
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert subzone.checkRupturePlane()

    # Bio-Bio, Chile
    slab_params = {
        "strike": 18.470233917236328,
        "dip": 20.36998748779297,
        "depth": 31.825963973999023,
    }
    tensor_params = {
        "NP1": {"strike": 178, "dip": 77, "rake": 86},
        "NP2": {"strike": 17, "dip": 14, "rake": 108},
        "P": {"azimuth": 272, "plunge": 32},
    }

    depth = 22.9
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert subzone.checkRupturePlane()

    # Sumatra
    slab_params = {
        "strike": 307.0806884765625,
        "dip": 20.450008392333984,
        "depth": 38.658607482910156,
    }
    tensor_params = {
        "NP1": {"strike": 336, "dip": 7, "rake": 114},
        "NP2": {"strike": 132, "dip": 84, "rake": 87},
        "P": {"azimuth": 224, "plunge": 39},
    }

    depth = 30.0
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert subzone.checkRupturePlane()

    # Sumatra, random moment tensor (should fail)
    slab_params = {
        "strike": 307.0806884765625,
        "dip": 20.450008392333984,
        "depth": 38.658607482910156,
    }
    tensor_params = {
        "NP1": {"strike": 224, "dip": 81, "rake": -169},
        "NP2": {"strike": 132, "dip": 79, "rake": -9},
        "P": {"azimuth": 88, "plunge": 14},
    }

    depth = 30.0
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert not subzone.checkRupturePlane()


def test_interface_depth():
    config = {
        "CONSTANTS": {
            "dstrike_interf": 30,
            "ddip_interf": 30,
            "dlambda": 60,
            "ddepth_interf": 20,
            "ddepth_intra": 10,
        }
    }
    # using slab information from tohoku epicenter
    # and MT info from duputel solution for tohoku
    slab_params = {
        "strike": 190.22474670410156,
        "dip": 14.894770622253418,
        "depth": 30.23206329345703,
    }

    tensor_params = {
        "NP1": {"strike": 193, "dip": 9, "rake": 78},
        "NP2": {"strike": 25, "dip": 81, "rake": 92},
        "P": {"azimuth": 113, "plunge": 36},
    }

    depth = 29.0
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert subzone.checkInterfaceDepth()

    # go down in the slab
    depth = 150
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert not subzone.checkInterfaceDepth()


def test_slab_depth():
    config = {
        "CONSTANTS": {
            "dstrike_interf": 30,
            "ddip_interf": 30,
            "dlambda": 60,
            "ddepth_interf": 20,
            "ddepth_intra": 10,
        }
    }
    # using slab information from tohoku epicenter
    # and MT info from duputel solution for tohoku
    slab_params = {
        "strike": 190.22474670410156,
        "dip": 14.894770622253418,
        "depth": 30.23206329345703,
    }

    tensor_params = {
        "NP1": {"strike": 193, "dip": 9, "rake": 78},
        "NP2": {"strike": 25, "dip": 81, "rake": 92},
        "P": {"azimuth": 113, "plunge": 36},
    }

    depth = 29.0
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert not subzone.checkSlabDepth(70.0)

    # go down in the slab
    depth = 150
    subzone = SubductionZone(slab_params, tensor_params, depth, config)
    assert subzone.checkSlabDepth(70.0)


if __name__ == "__main__":
    test_norm_angle()
    test_rupture_plane()
    test_interface_depth()
    test_slab_depth()



#!/usr/bin/env python
# stdlib imports
import pathlib

# third party imports
import numpy as np

# local imports
from strec.cmt import getCompositeCMT
from strec.subtype import get_focal_mechanism


def test_composite():
    homedir = pathlib.Path(__file__).parent  # where is this script?
    dbfile = homedir / ".." / "data" / "strec.db"  # ends 2016-10-31
    lat, lon, depth, _ = 3.295, 95.982, 30.0, 9.1  # sumatra
    tensor_params1, similarity, N = getCompositeCMT(
        lat, lon, dbfile, box=0.5, nmin=3.0, maxbox=1.0, dbox=0.1
    )

    testout = {
        "T": {
            "value": 0.90040807112868482,
            "azimuth": 59.257592723523928,
            "plunge": 60.956153671192382,
        },
        "type": "unknown",
        "NP1": {
            "rake": 99.706221357007934,
            "dip": 72.880835583169542,
            "strike": 135.04248070294602,
        },
        "source": "unknown",
        "N": {
            "value": 0.19843836648633859,
            "azimuth": 312.16014237459842,
            "plunge": 9.2722987641241801,
        },
        "P": {
            "value": -1.0987962433960725,
            "azimuth": 217.33561087867673,
            "plunge": 27.25558260027913,
        },
        "mrt": 0.5722075571696541,
        "mtp": 0.42166052289847422,
        "mpp": -0.056425346237488162,
        "mrr": 0.46289484505771517,
        "mtt": -0.40641930460127595,
        "NP2": {
            "rake": 61.304600053275323,
            "dip": 19.607420050830289,
            "strike": 284.8827314746506,
        },
        "mrp": -0.57636593662837132,
    }

    mechanism = get_focal_mechanism(tensor_params1)

    print("Testing that CMT composite tensor is consistent with past results...")
    assert mechanism == "RS"
    for key, value in tensor_params1.items():
        value1 = testout[key]
        if isinstance(value, float):
            np.testing.assert_almost_equal(value, value1)
        elif isinstance(value, str):
            assert value == value1
        else:
            for key2, value2 in value.items():
                value3 = value1[key2]
                try:
                    np.testing.assert_almost_equal(value2, value3)
                except Exception as e:
                    print(key2, value2, value3)
                    raise e

    np.testing.assert_almost_equal(similarity, 1.1036343285450121)
    assert N == 50
    print("Passed.")


if __name__ == "__main__":
    test_composite()



# stdlib imports
from enum import Enum

# third party imports
import numpy as np


class Zone(Enum):
    A = "A"
    B = "B"
    C = "C"


# This code is developed from the paper
# "NGA-Sub source and path database", Contreras et. al., 2022, Earthquake Spectra
# DOI: 10.1177/87552930211065054
# While the authors are presenting a database, not an algorithm, we took the
# description of the database creation-process as closely as possible.
# Note that the authors say "Following these initial classifications,
# event types were checked based on human interpretation."
# At the time of this writing, the differences between the results of this
# implementation of the algorithm and the database created by the authors are great
# enough that we decided it would be less confusing to output these results in the code.
# We will likely revive and modify this function at some point in the future.


class Category(Enum):
    INTERFACE = "interface"
    SLAB = "slab"
    CRUSTAL = "crustal"
    OUTER_RISE = "outer rise"
    UNDETERMINED = "undetermined"
    CRUSTAL_UNCERTAIN = "crustal uncertain"
    INTERFACE_UNCERTAIN = "interface uncertain"
    SLAB_UNCERTAIN = "slab uncertain"
    OUTER_RISE_UNCERTAIN = "outer rise uncertain"


class Focal(Enum):
    RS = "RS"
    SS = "SS"
    NM = "NM"
    ALL = "ALL"


BOUNDARY_DELTA = 5
DIST_THRESHOLD = 10
MIN_SLAB_DEPTH = 10
MAX_INTERFACE_DEPTH = 60
MAX_CRUSTAL_DEPTH = 30
SUBDUCTION_REGION = "Subduction"


def nga_categorize(eqinfo, input_strec_info):
    """Assign subduction zone categories according to Contreras et al 2022.

    DOI for full paper:
    DOI: 10.1177/87552930211065054

    This function assign strings where the authors assign numerical codes for category.

    0: interface
    1: slab
    2: crustal
    4: outer rise
    -999: undetermined
    -888: interface uncertain
    -777: slab uncertain
    -666: crustal uncertain
    -444: outer rise uncertain

    Args:
        eqinfo (list): [latitude, longitude, depth]
        input_strec_info (pandas Series): Pandas Series with results from
                                          getSubductionType().
    Returns:
        Input Series with added fields NGACategory and NGAHorizontalType.
    """
    strec_info = input_strec_info.copy()
    category_distance = 1e10
    if strec_info["TectonicRegion"] != SUBDUCTION_REGION:
        strec_info["NGACategory"] = "undetermined"
        return strec_info
    if np.isnan(strec_info["SlabModelDepth"]):
        strec_info["NGACategory"] = "undetermined"
        return strec_info
    _, _, depth = eqinfo
    horizontal_type = Zone.A
    if strec_info["SlabModelDepth"] > MIN_SLAB_DEPTH:
        horizontal_type = Zone.B
    if strec_info["SlabModelDepth"] > strec_info["SlabModelMaximumDepth"]:
        horizontal_type = Zone.C

    shallowest_interface = (
        strec_info["SlabModelDepth"] - strec_info["SlabModelDepthUncertainty"]
    )
    category = None
    if horizontal_type == Zone.A:
        category_distance = min(category_distance, abs(depth - MAX_INTERFACE_DEPTH))
        if depth > MAX_INTERFACE_DEPTH:
            category = Category.SLAB
        else:
            category = Category.OUTER_RISE
    elif horizontal_type == Zone.B:
        category_distance = min(category_distance, abs(depth - MAX_INTERFACE_DEPTH))
        category_distance = min(category_distance, abs(depth - shallowest_interface))
        if depth > MAX_INTERFACE_DEPTH:
            category = Category.SLAB
        elif depth < shallowest_interface:
            category = Category.CRUSTAL
        else:
            category = Category.INTERFACE
    else:  # zone C
        category_distance = min(category_distance, abs(depth - MAX_CRUSTAL_DEPTH))
        category_distance = min(category_distance, abs(depth - shallowest_interface))
        if depth < MAX_CRUSTAL_DEPTH:
            category = Category.CRUSTAL
        elif depth < shallowest_interface:
            category = Category.UNDETERMINED  # -999
        else:
            category = Category.SLAB

    delta_depth = abs(depth - strec_info["SlabModelDepth"])
    delta_max = depth - strec_info["SlabModelMaximumDepth"]
    if (
        category == Category.INTERFACE
        and strec_info["FocalMechanism"] != Focal.RS.value
        and delta_depth < DIST_THRESHOLD
    ):
        category_distance = min(category_distance, abs(delta_depth - DIST_THRESHOLD))
        category = Category.CRUSTAL

    # we are applying this to all horizontal categories, however it is not clear from
    # paper if this is intended.
    slab_or_rise = (
        category == Category.SLAB
        or category == Category.OUTER_RISE
        or category == Category.SLAB_UNCERTAIN
        or category == Category.OUTER_RISE_UNCERTAIN
    )
    if (
        slab_or_rise
        and strec_info["FocalMechanism"] == Focal.RS.value
        and delta_max < DIST_THRESHOLD
    ):
        category_distance = min(category_distance, abs(delta_max - DIST_THRESHOLD))
        category = Category.INTERFACE

    # this criteria is somewhat unclear from paper
    boundary_uncertain = category_distance < BOUNDARY_DELTA
    if category == Category.OUTER_RISE and (
        strec_info["FocalMechanism"] != Focal.NM.value or boundary_uncertain
    ):
        category = Category.OUTER_RISE_UNCERTAIN

    if category == Category.INTERFACE and (
        strec_info["FocalMechanism"] != Focal.RS.value or boundary_uncertain
    ):
        category = Category.INTERFACE_UNCERTAIN

    if category == Category.SLAB and (
        strec_info["FocalMechanism"] != Focal.NM.value or boundary_uncertain
    ):
        category = Category.SLAB_UNCERTAIN

    if category == Category.CRUSTAL and boundary_uncertain:
        category = Category.CRUSTAL_UNCERTAIN

    strec_info["NGACategory"] = category.value

    return strec_info



#!/usr/bin/env python

# stdlib imports
import pathlib

# third party imports
import numpy as np
import pandas as pd
from esi_utils_rupture.tensor import fill_tensor_from_angles

# local imports
from strec.ngasub import nga_categorize
from strec.subtype import SubductionSelector


def test_nga_categorize():
    datafile = pathlib.Path(__file__).parent / "data" / "big_focals.xlsx"
    dataframe = pd.read_excel(datafile)
    dataframe = dataframe.drop_duplicates(subset=["EventID"])
    subset_file = pathlib.Path(__file__).parent / "data" / "subsampled_nga_results.xlsx"
    subset = pd.read_excel(subset_file)
    idx = dataframe["EventID"].isin(subset["EventID"])
    dataframe = dataframe.loc[idx]
    subset = subset.sort_values(by="EventID")
    dataframe = dataframe.sort_values(by="EventID")
    selector = SubductionSelector(verbose=False, prefix="")
    results = []
    ic = 0
    for idx, row in dataframe.iterrows():
        # event_id = row["comcat_id"]
        # detail = get_event_details(event_id)
        # lat, lon, depth = detail["latitude"], detail["longitude"], detail["depth"]
        lat = row["EventLatitude"]
        lon = row["EventLongitude"]
        depth = row["EventDepth"]
        strike = row["NodalPlane1Strike"]
        dip = row["NodalPlane1Dip"]
        rake = row["NodalPlane1Rake"]
        eventid = row["EventID"]
        mag = row["EventMagnitude"]
        tensor_params = None
        if not np.isnan(strike):
            tensor_params = fill_tensor_from_angles(strike, dip, rake, mag)
        strec_info = selector.getSubductionType(
            lat,
            lon,
            depth,
            eventid=eventid,
            tensor_params=tensor_params,
        )
        nga_strec_info = nga_categorize([lat, lon, depth], strec_info)
        nga_strec_info["EventID"] = eventid
        results.append(nga_strec_info)
        ic += 1
        print(f"Processed event {ic} of {len(dataframe)} events.")
    results_frame = pd.DataFrame(results)
    assert results_frame["NGACategory"].equals(subset["NGACategory"])


if __name__ == "__main__":
    test_nga_categorize()



#!/usr/bin/env python

# stdlib imports
import logging
import pathlib

# third party imports
import numpy as np
import requests
from esi_utils_rupture.tensor import fill_tensor_from_components

# local imports
from strec.cmt import getCompositeCMT
from strec.gmreg import Regionalizer
from strec.kagan import get_kagan_angle
from strec.slab import SlabCollection
from strec.utils import get_config

EVENT_URL = (
    "https://earthquake.usgs.gov/fdsnws/event/1/query?eventid=EVENTID&format=geojson"
)
SLAB_RAKE = 90  # presumed rake angle of slabs

SLAB_REGIONS = {
    "alu": "Alaska-Aleutians",
    "cal": "Calabria",
    "cam": "Central America",
    "car": "Caribbean",
    "cas": "Cascadia",
    "cot": "Cotabato",
    "hal": "Halmahera",
    "hel": "Helanic",
    "him": "Himalaya",
    "hin": "Hindu Kush",
    "izu": "Izu-Bonin",
    "ker": "Kermadec-Tonga",
    "kur": "Kamchatka/Kurils/Japan",
    "mak": "Makran",
    "man": "Manila",
    "mex": "Central America",
    "mue": "Muertos",
    "pam": "Pamir",
    "pan": "Panama",
    "phi": "Philippines",
    "png": "New Guinea",
    "puy": "Puysegur",
    "ryu": "Ryukyu",
    "sam": "South America",
    "sco": "Scotia",
    "sol": "Solomon Islands",
    "sul": "Sulawesi",
    "sum": "Sumatra-Java",
    "van": "Santa Cruz Islands/Vanuatu/Loyalty Islands",
}

CONSTANTS = {
    "tplunge_rs": 50,
    "bplunge_ds": 30,
    "bplunge_ss": 55,
    "pplunge_nm": 55,
    "delplunge_ss": 20,
}

COMCAT_TEMPLATE = (
    "https://earthquake.usgs.gov/earthquakes/feed/v1.0/detail/[EVENTID].geojson"
)


def fill_axis(axis, tensor_props):
    axis_dict = {}
    if f"{axis.lower()}-axis-plunge" in tensor_props:
        axis_dict[f"{axis.upper()}"] = {
            "azimuth": float(tensor_props[f"{axis.lower()}-axis-azimuth"]),
            "plunge": float(tensor_props[f"{axis.lower()}-axis-plunge"]),
        }
    return axis_dict


def fill_nodal_plane(nodal_plane, tensor_props):
    # nodal_plane = NP1/2
    planes = {
        "NP1": "nodal-plane-1",
        "NP2": "nodal-plane-2",
    }
    plane = planes[nodal_plane]
    nodal_dict = {}
    rake_str = f"{plane}-rake"
    slip_str = f"{plane}-slip"
    strike_str = f"{plane}-strike"
    dip_str = f"{plane}-dip"
    if rake_str in tensor_props:
        nodal_dict["rake"] = float(tensor_props[rake_str])
    else:
        nodal_dict["rake"] = float(tensor_props[slip_str])
    nodal_dict["strike"] = float(tensor_props[strike_str])
    nodal_dict["dip"] = float(tensor_props[dip_str])
    return nodal_dict


def get_event_details(eventid):
    event_dict = {}
    url = COMCAT_TEMPLATE.replace("[EVENTID]", eventid)
    response = requests.get(url)
    jdict = response.json()
    event_dict["latitude"] = jdict["geometry"]["coordinates"][1]
    event_dict["longitude"] = jdict["geometry"]["coordinates"][0]
    event_dict["depth"] = jdict["geometry"]["coordinates"][2]
    event_dict["magnitude"] = jdict["properties"]["mag"]
    mech_type = "moment-tensor"
    if "moment-tensor" not in jdict["properties"]["products"].keys():
        if "focal-mechanism" not in jdict["properties"]["products"].keys():
            return event_dict
        mech_type = "focal-mechanism"
    tensor_dict = jdict["properties"]["products"][mech_type][0]
    tensor_props = tensor_dict["properties"]
    tensor_type = tensor_props.get("derived-magnitude-type", "unknown")
    if tensor_type == "unknown":
        btype = tensor_props.get("beachball-type", "unknown")
        if btype != "unknown":
            if btype.find("/") > -1:
                btype = btype.split("/")[-1]
            tensor_type = btype
    tensor = {}
    tensor["type"] = tensor_type
    tensor["source"] = "unknown"
    tensor["source"] = tensor_props.get("eventsource", "unknown")
    if tensor["source"] == "unknown":
        tensor["source"] = tensor_props.get("beachball-source", "unknown")
    if mech_type == "moment-tensor":
        tensor["mrr"] = float(tensor_props["tensor-mrr"])
        tensor["mtt"] = float(tensor_props["tensor-mtt"])
        tensor["mpp"] = float(tensor_props["tensor-mpp"])
        tensor["mrt"] = float(tensor_props["tensor-mrt"])
        tensor["mrp"] = float(tensor_props["tensor-mrp"])
        tensor["mtp"] = float(tensor_props["tensor-mtp"])
    for axis in ["T", "N", "P"]:
        axis_dict = fill_axis(axis, tensor_props)
        tensor.update(axis_dict)
    for nodal_plane in ["NP1", "NP2"]:
        if "nodal-plane-1-dip" not in tensor_props:
            continue
        tensor[nodal_plane] = fill_nodal_plane(nodal_plane, tensor_props)
    event_dict["tensor"] = tensor
    return event_dict


class SubductionSelector(object):
    """For subduction events, determine subduction zone properties."""

    def __init__(self, prefix=None, verbose=False):
        """Construct a SubductionSelector object."""
        if prefix is not None:
            self.logger = logging.getLogger(prefix)
        else:
            self.logger = logging.getLogger()
        self.verbose = verbose
        self._regionalizer = Regionalizer.load()
        self._config = get_config()

    def getSubductionTypeByID(self, eventid):
        """Given an event ID, determine the subduction zone information.

        Args:
            eventid (str): ComCat EventID (Sumatra is official20041226005853450_30).
        Returns:
            Pandas Series object with indices:
                - TectonicRegion : (Subduction,Active,Stable,HotSpot)
                - TectonicDomain : SZ (generic)
                - FocalMechanism : (RS [Reverse],SS [Strike-Slip], NM [Normal], ALL
                [Unknown])
                - TensorType : (actual, composite)
                - KaganAngle : Angle between moment tensor and slab interface.
                - DistanceToStable : Distance in km from the nearest stable polygon.
                - DistanceToActive : Distance in km from the nearest active polygon.
                - DistanceToSubduction : Distance in km from the nearest subduction
                polygon.
                - DistanceToHotSpot : Distance in km from the nearest hotspot polygon.
                - Oceanic : Boolean indicating whether we are in an oceanic region.
                - DistanceToOceanic : Distance in km to nearest oceanic polygon.
                - DistanceToContinental : Distance in km to nearest continental polygon.
                - TectonicSubtype : (SZInter,ACR,SZIntra)
                - RegionContainsBackArc : Boolean indicating whether event is in a
                back-arc subduction region.
                - DomainDepthBand1 : Bottom of generic depth level for shallowest
                subduction type.
                - DomainDepthBand1Subtype : Shallowest subduction type.
                - DomainDepthBand2 : Bottom of generic depth level for middle
                subduction type.
                - DomainDepthBand2Subtype : Middle subduction type.
                - DomainDepthBand3 : Bottom of generic depth level for deepest
                subduction type.
                - DomainDepthBand3Subtype : Deepest subduction type
                - SlabModelRegion : Subduction region.
                - SlabModelType : (grid,trench)
                - SlabModelDepth : Depth to slab interface at epicenter.
                - SlabModelDip : Dip of slab at epicenter.
                - SlabModelStrike : Strike of slab at epicenter.
                - IsLikeInterface : Boolean indicating whether moment tensor strike is
                similar to interface.
                - IsNearInterface : Boolean indicating whether depth is close to
                interface.
                - IsInSlab : Boolean indicating whether depth is within the slab.
        Raises:
            AttributeError if the eventid is not found in ComCat.
        """
        if self.verbose:
            self.logger.info("Inside getSubductionTypeByID...")
        lat, lon, depth, tensor_params = self.getOnlineTensor(eventid)
        if self.verbose:
            self.logger.info("Tensor Parameters: %s" % str(tensor_params))
        if lat is None:
            raise AttributeError("Event %s is not found in ComCat." % eventid)

        lat = float(lat)
        lon = float(lon)
        results = self.getSubductionType(lat, lon, depth, tensor_params=tensor_params)
        return results

    def getOnlineTensor(self, eventid):
        """Get tensor parameters from preferred ComCat moment tensor.

        Args:
            eventid (str): ComCat EventID (Sumatra is official20041226005853450_30).
        Returns:
            Moment tensor parameters dictionary:
                - source Moment Tensor source
                - type usually mww,mwc,mwb,mwr,TMTS or "unknown".
                - mrr,mtt,mpp,mrt,mrp,mtp Moment tensor components.
                - T T-axis values:
                  - azimuth
                  - plunge
                - N N-axis values:
                  - azimuth
                  - plunge
                - P P-axis values:
                  - azimuth
                  - plunge
                - NP1 First nodal plane values:
                  - strike
                  - dip
                  - rake
                - NP2 Second nodal plane values:
                  - strike
                  - dip
                  - rake
        """
        if self.verbose:
            self.logger.info("Inside getOnlineTensor")
        try:
            detail = get_event_details(eventid)
        except Exception as e:
            msg = 'Failed to get event information for %s - error "%s"'
            tpl = (eventid, str(e))
            self.logger.warn(msg % tpl)
            return (None, None, None, None)
        lat = detail["latitude"]
        lon = detail["longitude"]
        depth = detail["depth"]
        if "tensor" not in detail:
            self.logger.info("No moment tensor available for %s" % eventid)
            return lat, lon, depth, None

        if self.verbose:
            self.logger.info("Getting tensor components...")
        tensor_params = detail["tensor"]

        # if no tensor, bail out
        if tensor_params is None:
            lat, lon, depth, tensor_params

        if self.verbose:
            self.logger.info("Getting tensor axes...")
        # sometimes the online MT is missing properties

        if "T" not in tensor_params and "mrr" in tensor_params:
            if self.verbose:
                self.logger.info("Calling fill_tensor function...")
            tensor_dict = fill_tensor_from_components(
                tensor_params["mrr"],
                tensor_params["mtt"],
                tensor_params["mpp"],
                tensor_params["mrt"],
                tensor_params["mrp"],
                tensor_params["mtp"],
            )
            tensor_params["T"] = tensor_dict["T"].copy()
            tensor_params["N"] = tensor_dict["T"].copy()
            tensor_params["P"] = tensor_dict["P"].copy()

        if self.verbose:
            self.logger.info("Getting tensor angles...")
        if "NP1" not in tensor_params and "mrr" in tensor_params:
            if self.verbose:
                self.logger.info("Calling fill_tensor function...")
            tensor2 = fill_tensor_from_components(
                tensor_params["mrr"],
                tensor_params["mtt"],
                tensor_params["mpp"],
                tensor_params["mrt"],
                tensor_params["mrp"],
                tensor_params["mtp"],
            )
            tensor_params["NP1"] = tensor2["NP1"].copy()
            tensor_params["NP2"] = tensor2["NP2"].copy()

        return lat, lon, depth, tensor_params

    def getSubductionType(self, lat, lon, depth, eventid=None, tensor_params=None):
        """Given a event hypocenter, determine the subduction zone information.

        Args:
            lat (float): Epicentral latitude.
            lon (float): Epicentral longitude.
            depth (float): Epicentral depth.
            eventid (float): ComCat EventID (Sumatra is official20041226005853450_30).
            tensor_params (dict): Dictionary containing moment tensor parameters:
                - mrr,mtt,mpp,mrt,mrp,mtp Moment tensor components.
                - T T-axis values:
                  - azimuth
                  - plunge
                - N N-axis values:
                  - azimuth
                  - plunge
                - P P-axis values:
                  - azimuth
                  - plunge
                - NP1 First nodal plane values:
                  - strike
                  - dip
                  - rake
                - NP2 Second nodal plane values:
                  - strike
                  - dip
                  - rake
                (optional) - type Moment Tensor type.
                (optional) - source Moment Tensor source (regional network, name of
                study, etc.)
        Returns:
            Pandas Series object with indices:
                - TectonicRegion : (Subduction,Active,Stable,HotSpot)
                - FocalMechanism : (RS [Reverse],SS [Strike-Slip], NM [Normal], ALL
                [Unknown])
                - TensorType : (actual, composite)
                - TensorSource : String indicating the source of the moment tensor
                information.
                - KaganAngle : Angle between moment tensor and slab interface.
                - CompositeVariability : A measure of the uncertainty in the composite
                moment tensor.
                - NComposite : Number of events used to create composite moment tensor.
                - DistanceToStable : Distance in km from the nearest stable polygon.
                - DistanceToActive : Distance in km from the nearest active polygon.
                - DistanceToSubduction : Distance in km from the nearest subduction
                polygon.
                - DistanceToHotSpot : Distance in km from the nearest hotspot polygon.
                - Oceanic : Boolean indicating whether we are in an oceanic region.
                - DistanceToOceanic : Distance in km to nearest oceanic polygon.
                - DistanceToContinental : Distance in km to nearest continental polygon.
                - SlabModelRegion : Subduction region.
                - SlabModelType : (grid,trench)
                - SlabModelDepth : Depth to slab interface at epicenter.
                - SlabModelDepthUncertainty : Uncertainty of depth to slab interface.
                - SlabModelDip : Dip of slab at epicenter.
                - SlabModelStrike : Strike of slab at epicenter.
        """
        if self.verbose:
            self.logger.info("Inside getSubductionType...")
        # sometimes events are specified with negative depths, which don't work with
        # our algorithms.  Pin those depths to 0.
        if depth < 0:
            depth = 0

        config = self._config
        slab_data_folder = pathlib.Path(config["DATA"]["slabfolder"])
        data_folder = pathlib.Path(config["DATA"]["folder"])
        tensor_type = None
        tensor_source = None
        similarity = np.nan
        nevents = 0
        if tensor_params is None:
            if eventid is not None:
                _, _, _, tensor_params = self.getOnlineTensor(eventid)
                if tensor_params is not None:
                    tensor_type = tensor_params["type"]
                    tensor_source = tensor_params["source"]

            if tensor_params is None:
                dbfile = data_folder / config["DATA"]["dbfile"]
                minboxcomp = float(config["CONSTANTS"]["minradial_distcomp"])
                maxboxcomp = float(config["CONSTANTS"]["maxradial_distcomp"])
                dboxcomp = float(config["CONSTANTS"]["step_distcomp"])

                # Minimum number of events required to get composite mechanism
                nmin = int(config["CONSTANTS"]["minno_comp"])
                tensor_params, similarity, nevents = getCompositeCMT(
                    lat,
                    lon,
                    dbfile,
                    box=minboxcomp,
                    maxbox=maxboxcomp,
                    dbox=dboxcomp,
                    nmin=nmin,
                )
                if tensor_params is not None:
                    tensor_type = "composite"
                    tensor_source = "composite"
        else:
            if "type" in tensor_params:
                tensor_type = tensor_params["type"]
            if "source" in tensor_params:
                tensor_source = tensor_params["source"]

        slab_collection = SlabCollection(slab_data_folder)
        slab_params = slab_collection.getSlabInfo(lat, lon, depth)

        results = self._regionalizer.getRegions(lat, lon)
        results["TensorType"] = tensor_type
        results["TensorSource"] = tensor_source
        results["CompositeVariability"] = similarity
        results["NComposite"] = nevents
        results["FocalMechanism"] = get_focal_mechanism(tensor_params)
        if len(slab_params):
            if np.isnan(slab_params["depth"]):
                results["SlabModelRegion"] = SLAB_REGIONS[slab_params["region"]]
                results["KaganAngle"] = np.nan
            else:
                results["SlabModelRegion"] = SLAB_REGIONS[slab_params["region"]]
                results["SlabModelDepth"] = slab_params["depth"]
                results["SlabModelDepthUncertainty"] = slab_params["depth_uncertainty"]
                results["SlabModelDip"] = slab_params["dip"]
                results["SlabModelStrike"] = slab_params["strike"]
                results["SlabModelMaximumDepth"] = slab_params[
                    "maximum_interface_depth"
                ]
                if tensor_params is not None:
                    np1 = tensor_params["NP1"]
                    kagan = get_kagan_angle(
                        slab_params["strike"],
                        slab_params["dip"],
                        SLAB_RAKE,
                        np1["strike"],
                        np1["dip"],
                        np1["rake"],
                    )
                    results["KaganAngle"] = kagan
                else:
                    results["KaganAngle"] = np.nan
        else:
            results["SlabModelRegion"] = ""
            results["SlabModelDepth"] = np.nan
            results["SlabModelDepthUncertainty"] = np.nan
            results["SlabModelDip"] = np.nan
            results["SlabModelStrike"] = np.nan
            results["SlabModelMaximumDepth"] = np.nan
            results["KaganAngle"] = np.nan

        results = results.reindex(
            index=[
                "TectonicRegion",
                "FocalMechanism",
                "TensorType",
                "TensorSource",
                "KaganAngle",
                "CompositeVariability",
                "NComposite",
                "DistanceToStable",
                "DistanceToActive",
                "DistanceToSubduction",
                "DistanceToHotSpot",
                "Oceanic",
                "DistanceToOceanic",
                "DistanceToContinental",
                "SlabModelRegion",
                "SlabModelDepth",
                "SlabModelDepthUncertainty",
                "SlabModelDip",
                "SlabModelStrike",
                "SlabModelMaximumDepth",
            ]
        )

        return results


def get_focal_mechanism(tensor_params):
    """Return focal mechanism (strike-slip,normal, or reverse).

    Args:
        tensor_params (dict): Dictionary containing the following fields:
            - 'T' Dictionary of 'azimuth' and 'plunge' values for the T axis.
            - 'N' Dictionary of 'azimuth' and 'plunge' values for the N(B) axis.
            - 'P' Dictionary of 'azimuth' and 'plunge' values for the P axis.
            - 'NP1' Dictionary of angles for the first nodal plane ('strike','dip',
            'rake')
            - 'NP2' Dictionary of angles for the second nodal plane ('strike','dip',
            'rake')
        config (dict): dictionary containing:
            - constants:
            - tplunge_rs
            - bplunge_ds
            - bplunge_ss
            - pplunge_nm
            - delplunge_ss
    Returns:
        str: Focal mechanism string 'SS','RS','NM',or 'ALL'.
    """
    if tensor_params is None or "T" not in tensor_params:
        return "ALL"
    # implement eq 1 here
    Tp = tensor_params["T"]["plunge"]
    Np = tensor_params["N"]["plunge"]
    Pp = tensor_params["P"]["plunge"]
    tplunge_rs = CONSTANTS["tplunge_rs"]
    bplunge_ds = CONSTANTS["bplunge_ds"]
    bplunge_ss = CONSTANTS["bplunge_ss"]
    pplunge_nm = CONSTANTS["pplunge_nm"]
    delplunge_ss = CONSTANTS["delplunge_ss"]
    if Tp >= tplunge_rs and Np <= bplunge_ds:
        return "RS"
    if Np >= bplunge_ss and (Tp >= Pp - delplunge_ss and Tp <= Pp + delplunge_ss):
        return "SS"
    if Pp >= pplunge_nm and Np <= bplunge_ds:
        return "NM"
    return "ALL"



#!/usr/bin/env python

# stdlib imports
import pathlib

# third party imports
from configobj import ConfigObj

from strec.sm_probs import get_probs
from strec.subtype import SubductionSelector, get_event_details


def _test_get_region_probs():
    select_file = pathlib.Path.home() / ".strec" / "select.conf"
    config = ConfigObj(str(select_file))
    eventid = "ak0219neiszm"
    selector = SubductionSelector(verbose=False, prefix="")
    lat, lon, depth, tensor_params = selector.getOnlineTensor(eventid)
    details = get_event_details(eventid)
    strec_info = selector.getSubductionTypeByID(eventid)
    probs = get_probs(details["magnitude"], depth, strec_info, config)
    x = 1


if __name__ == "__main__":
    test_get_region_probs()



# stlib imports
import logging
import re

# third party imports
import numpy as np


def get_config_regions(config):
    tectonic_regions = {}
    for section, options in config.items():
        if not section.endswith("REGION"):
            continue
        region = section.replace("_REGION", "").lower()
        rdict = {}
        for key, value in options.items():
            if value == "True":
                rdict[key] = bool(value)
            else:
                values = [float(v) for v in value.split(",")]
                if len(values) > 1 or "depth" in key:
                    rdict[key] = values
                else:
                    rdict[key] = values[0]
        tectonic_regions[region] = rdict
    return tectonic_regions


def fix_config(config):
    for key, value in config.items():
        if key in ["min_depth", "max_depth"] and not isinstance(value, list):
            value = [value]
        if isinstance(value, str):
            try:
                value = float(value)
                config[key] = value
            except ValueError:
                continue
        if isinstance(value, list):
            tmpvalue = []
            for v in value:
                try:
                    tmpv = float(v)
                    tmpvalue.append(tmpv)
                except ValueError:
                    tmpvalue.append(v)
            config[key] = tmpvalue
        if isinstance(value, dict):
            fix_config(value)


def get_probs(mag, depth, strec_results, config):
    fix_config(config)
    region_probs = get_region_probs(depth, strec_results, config)
    in_subduction = strec_results["TectonicRegion"] == "Subduction"
    above_slab = not np.isnan(strec_results["SlabModelDepth"])

    use_slab = config["tectonic_regions"]["subduction"]["use_slab"]

    if use_slab:
        if in_subduction:
            subduction_probs = get_subduction_probs(
                strec_results, depth, mag, config["subduction"], above_slab
            )
            for key, value in subduction_probs.items():
                subduction_probs[key] = value * region_probs["subduction"]

            # If we are in a subduction zone then we don't want the
            # keys for subduction_0, 1, 2 (which are the generic vertical
            # subduction subtypes that are not informed by the slab model because
            # it isn't available)
            if "subduction_0" in region_probs:
                del region_probs["subduction_0"]
            if "subduction_1" in region_probs:
                del region_probs["subduction_1"]
            if "subduction_2" in region_probs:
                del region_probs["subduction_2"]

        else:
            # If we are NOT in a subduction zone we may or may not need subduction
            # probabilities (depending on distance and the configured taper). But
            # either way, we will not have access to the slab model and so we have
            # to use the generic vertical subtypes
            subduction_probs = {
                "crustal": region_probs["subduction_0"],
                "interface": region_probs["subduction_1"],
                "intraslab": region_probs["subduction_2"],
            }
        region_probs.update(subduction_probs)
    else:
        logging.info('"use_slab" is False so no slab used in finding GMPE ' "weights.")

    return region_probs


def get_region_probs(depth, strec_results, config):
    """
    Calculate the regional probabilities (not including subduction interface
    etc).

    Args:
        depth (float): Depth of earthquake.
        strec_results (Series): Pandas series containing STREC output.
        config (dict-like): Configuration information regarding earthquake
            type.

    Returns:
        dict: Probabilities for each earthquake type, with fields:
              - acr Probability that the earthquake is in an active region.
              - acr_X Probability that the earthquake is in a depth layer of
                ACR, starting from the top.
              - scr Probability that the earthquake is in a stable region.
              - scr_X Probability that the earthquake is in a depth layer of
                SCR, starting from the top.
              - volcanic Probability that the earthquake is in a volcanic
                region.
              - volcanic_X Probability that the earthquake is in a depth layer
                of Volcanic, starting from the top.
              - subduction Probability that the earthquake is in a subduction
                zone.

    """
    tectonic_regions = config["tectonic_regions"]
    region_probs = {}
    region_mapping = {
        "scr": "DistanceToStable",
        "acr": "DistanceToActive",
        "volcanic": "DistanceToHotSpot",
        "subduction": "DistanceToSubduction",
    }
    layer_probs = {}

    for region, rdict in tectonic_regions.items():
        distance = strec_results[region_mapping[region]]
        # If we're considering subduction zones but not IN a subduction zone
        x1 = 0.0
        p2 = 0.0

        p1 = 1.0
        x2 = rdict["horizontal_buffer"]

        region_prob = get_probability(distance, x1, p1, x2, p2)
        region_probs[region] = region_prob

        region_layer_probs = {}
        # now do the weights for each depth zone
        for i in range(0, len(rdict["min_depth"])):
            # First, taper from -1 to 0 for the lower end
            x1 = rdict["min_depth"][i] - rdict["vertical_buffer"] / 2
            p1 = -1.0
            x2 = rdict["min_depth"][i]
            p2 = 0.0
            p_layer1 = get_probability(depth, x1, p1, x2, p2)
            # Then, taper from 0 to -1 for the higher end
            x1 = rdict["max_depth"][i]
            p1 = 0.0
            x2 = rdict["max_depth"][i] + rdict["vertical_buffer"] / 2
            p2 = -1.0
            p_layer2 = get_probability(depth, x1, p1, x2, p2)
            # Lastly, combine to get probability curve for layer i
            region_layer_probs["%s_%i" % (region, i)] = 1 + p_layer1 + p_layer2
        probsum = sum([lp for lp in list(region_layer_probs.values())])
        if probsum > 0:
            for key, value in region_layer_probs.items():
                region_layer_probs[key] = value / probsum

        # running list of all region layer probabilities
        layer_probs.update(region_layer_probs)

    # divide the weights by the total weight
    probsum = sum(list(region_probs.values()))
    for region, prob in region_probs.items():
        region_probs[region] = prob / probsum
        pat = re.compile(region)
        layerkeys = list(layer_probs.keys())
        reg_layers = list(filter(pat.search, layerkeys))
        for reg_layer in reg_layers:
            layer_probs[reg_layer] = layer_probs[reg_layer] * region_probs[region]

    region_probs.update(layer_probs)

    return region_probs


def get_subduction_probs(strec_results, depth, mag, subcfg, above_slab):
    """Get probabilities of earthquake being crustal, interface or intraslab.

    Args:
        strec_results (Series): Pandas series containing STREC output.
        depth (float): Depth of earthquake.
        mag (float): Earthquake magnitude.
        config (dict-like): Configuration information regarding earthquake
            type.
        above_slab (bool): Is earthquake above a defined slab?

    Returns:
        dict: Probabilities for each earthquake type, with fields:
              - crustal Probability that the earthquake is in the crust above
                an interface.
              - interface Probability that the earthquake is on the interface.
              - intraslab Probability that the earthquake is in the slab below
                interface.

    """
    if above_slab:
        # the angle between moment tensor and slab
        kagan = strec_results["KaganAngle"]  # can be nan

        # Depth to slab
        slab_depth = strec_results["SlabModelDepth"]

        # Error in depth to slab
        slab_depth_error = strec_results["SlabModelDepthUncertainty"]

        # what is the effective bottom of the interface zone?
        max_interface_depth = strec_results["SlabModelMaximumDepth"]

        # Calculate the probability of interface given the
        # (absolute value of) difference between hypocenter and depth to slab.
        dz = np.abs(depth - slab_depth)

        x1 = subcfg["p_int_hypo"]["x1"] + slab_depth_error
        x2 = subcfg["p_int_hypo"]["x2"] + slab_depth_error
        p1 = subcfg["p_int_hypo"]["p1"]
        p2 = subcfg["p_int_hypo"]["p2"]
        p_int_hypo = get_probability(dz, x1, p1, x2, p2)

        # Calculate probability of interface given Kagan's angle
        if np.isfinite(kagan):
            x1 = subcfg["p_int_kagan"]["x1"]
            x2 = subcfg["p_int_kagan"]["x2"]
            p1 = subcfg["p_int_kagan"]["p1"]
            p2 = subcfg["p_int_kagan"]["p2"]
            p_int_kagan = get_probability(kagan, x1, p1, x2, p2)
        else:
            p_int_kagan = subcfg["p_kagan_default"]

        # Calculate probability that event occurred above bottom of seismogenic
        # zone, given to us by the Slab model.
        x1 = max_interface_depth + subcfg["p_int_sz"]["x1"]
        x2 = max_interface_depth + subcfg["p_int_sz"]["x2"]
        p1 = subcfg["p_int_sz"]["p1"]
        p2 = subcfg["p_int_sz"]["p2"]
        p_int_sz = get_probability(depth, x1, p1, x2, p2)

        # Calculate combined probability of interface
        p_int = p_int_hypo * p_int_kagan * p_int_sz

        # Calculate probability that the earthquake lies above the slab
        # and is thus crustal.
        x1 = subcfg["p_crust_slab"]["x1"]
        x2 = subcfg["p_crust_slab"]["x2"]
        p1 = subcfg["p_crust_slab"]["p1"]
        p2 = subcfg["p_crust_slab"]["p2"]

        p_crust_slab = get_probability((depth - slab_depth), x1, p1, x2, p2)

        # Calculate probability that the earthquake lies within the crust
        x1 = subcfg["p_crust_hypo"]["x1"]
        x2 = subcfg["p_crust_hypo"]["x2"]
        p1 = subcfg["p_crust_hypo"]["p1"]
        p2 = subcfg["p_crust_hypo"]["p2"]
        p_crust_hypo = get_probability(depth, x1, p1, x2, p2)

        # Calculate probability of crustal
        p_crustal = (1 - p_int) * p_crust_slab * p_crust_hypo

        # Calculate probability of intraslab
        p_slab = 1 - (p_int + p_crustal)

    else:
        slab_depth = subcfg["default_slab_depth"]
        # Calculate the probability that an earthquake is interface
        # given magnitude
        x1 = subcfg["p_int_mag"]["x1"]
        p1 = subcfg["p_int_mag"]["p1"]
        x2 = subcfg["p_int_mag"]["x2"]
        p2 = subcfg["p_int_mag"]["p2"]
        p_int_mag = get_probability(mag, x1, p1, x2, p2)

        # Calculate the probability that the earthquake is
        # interface given depth (two part function).
        # upper portion of function
        x1 = subcfg["p_int_dep_no_slab_upper"]["x1"]
        p1 = subcfg["p_int_dep_no_slab_upper"]["p1"]
        x2 = subcfg["p_int_dep_no_slab_upper"]["x2"]
        p2 = subcfg["p_int_dep_no_slab_upper"]["p2"]
        p_int_depth_upper = get_probability(depth, x1, p1, x2, p2)

        # lower portion of function
        x1 = subcfg["p_int_dep_no_slab_lower"]["x1"]
        p1 = subcfg["p_int_dep_no_slab_lower"]["p1"]
        x2 = subcfg["p_int_dep_no_slab_lower"]["x2"]
        p2 = subcfg["p_int_dep_no_slab_lower"]["p2"]
        p_int_depth_lower = get_probability(depth, x1, p1, x2, p2)

        p_int_depth = p_int_depth_upper + p_int_depth_lower

        # This functional form is used so that the probability of interface
        # inflates and appraoches 1 as magnitude gets large, assuming
        # that the ramp function for p_int_mag is zero at small magnitudes
        # and 1 at large magnitudes.
        p_int = p_int_depth + (1 - p_int_depth) * p_int_mag

        if depth > slab_depth:
            p_crustal = 0.0
            p_slab = 1 - p_int
        else:
            p_crustal = 1 - p_int
            p_slab = 0.0

    probs = {"crustal": p_crustal, "interface": p_int, "intraslab": p_slab}
    return probs


def get_probability(x, x1, p1, x2, p2):
    """Calculate probability using a ramped function.

    The subsections and parameters below reflect a series of ramp functions
    we use to calculate various probabilities.
    ::
      p1  |----+
          |     \
          |      \
          |       \
      p2  |        +-------
          |
          +-----------------
               x1  x2

    Args:
        x (float): Quantity for which we want corresponding probability.
        x1 (float): Minimum X value.
        p1 (float): Probability at or below minimum X value.
        x2 (float): Maximum X value.
        p2 (float): Probability at or below maximum X value.

    Returns:
        float: Probability at input x value.

    """
    if x <= x1:
        prob = p1
    elif x >= x2:
        prob = p2
    else:
        slope = (p1 - p2) / (x1 - x2)
        intercept = p1 - slope * x1
        prob = x * slope + intercept
    return prob



# stdlib imports
import os.path
import pathlib
import sqlite3
from collections import OrderedDict

# third party imports
import numpy as np
import pandas as pd

SCHEMA = OrderedDict(
    [
        ("time", "datetime"),
        ("sourceid", "integer"),
        ("lat", "float"),
        ("lon", "float"),
        ("depth", "float"),
        ("mag", "float"),
        ("mrr", "float"),
        ("mtt", "float"),
        ("mpp", "float"),
        ("mrt", "float"),
        ("mrp", "float"),
        ("mtp", "float"),
    ]
)

TIMEFMT = "%Y-%m-%d %H:%M:%S.%f"


def read_datafile(filename):
    """Read input moment data file (CSV/Excel) into dataframe.

    Args:
        filename (str): Excel/CSV file which must contain required columns (see above.)
    Returns:
        pd.DataFrame: Pandas dataframe representation of input file.
    """
    dataframe = None
    dtypes = {
        "mrr": np.float64,
        "mtt": np.float64,
        "mpp": np.float64,
        "mrt": np.float64,
        "mrp": np.float64,
        "mtp": np.float64,
    }
    try:
        dataframe = pd.read_csv(filename, dtype=dtypes)
    except Exception:
        try:
            dataframe = pd.read_excel(filename, dtype=dtypes)
        except Exception:
            raise Exception("Input file is not CSV or Excel.")
    req_columns = [
        "time",
        "lat",
        "lon",
        "depth",
        "mag",
        "mrr",
        "mtt",
        "mpp",
        "mrt",
        "mrp",
        "mtp",
    ]
    if not set(req_columns) <= set(dataframe.columns):
        missing = set(req_columns) - set(dataframe.columns)
        raise Exception("Missing columns: %s" % str(missing))

    # delete the columns that are no in required list
    for column in dataframe.columns:
        if column not in req_columns:
            dataframe = dataframe.drop(column, 1)

    return dataframe


def stash_dataframe(dataframe, datafile, source, create_db=False):
    """Store a dataframe in the database.

    Args:
        dataframe (DataFrame):
            pandas Dataframe, containing columns:
                - time (YYYY-MM-DD HH:MM:SS for CSV)
                - lat (decimal degrees)
                - lon (decimal degrees)
                - depth (km)
                - mag Magnitude
                - mrr Mrr moment tensor component
                - mtt Mtt moment tensor component
                - mpp Mpp moment tensor component
                - mrt Mrt moment tensor component
                - mrp Mrp moment tensor component
                - mtp Mtp moment tensor component
        datafile (str):
            Path to SQLite file where dataframe will be stored as a row.
        source (str):
            Network that contributed the data in the dataframe ("us","gcmt", etc.)
        create_db (bool):
            Boolean indicating whether to create a new database file or not.
    """
    datafile = pathlib.Path(datafile)
    if create_db:
        if datafile.is_file():
            os.remove(datafile)
        conn = sqlite3.connect(datafile)
        cursor = conn.cursor()
        nuggets = []
        for key, value in SCHEMA.items():
            nuggets.append("%s %s" % (key, value))
        create_stmt = "CREATE TABLE earthquake (%s)" % (",".join(nuggets))
        cursor.execute(create_stmt)
        source_stmt = "CREATE TABLE source (id integer primary key, source text)"
        cursor.execute(source_stmt)
    else:
        conn = sqlite3.connect(datafile)
        cursor = conn.cursor()

    cursor.execute('SELECT id from source where source = "%s"' % source.lower())
    sourcerow = cursor.fetchone()
    if sourcerow is not None:
        sourceid = sourcerow[0]
        dataframe["sourceid"] = sourceid
    else:
        cursor.execute('INSERT INTO source (source) values ("%s")' % source.lower())
        conn.commit()
        cursor.execute('SELECT id from source where source = "%s"' % source.lower())
        sourceid = cursor.fetchone()[0]
        dataframe["sourceid"] = sourceid

    dataframe.to_sql("earthquake", conn, if_exists="append", index=False)

    conn.close()


def fetch_dataframe(datafile):
    """Return a pandas dataframe containing earthquake information.

    Args:
        datafile (str):
            Path to sqlite3 database file.
    Returns:
      DataFrame:
        pandas Dataframe, containing columns:
            - time (YYYY-MM-DD HH:MM:SS for CSV)
            - lat (decimal degrees)
            - lon (decimal degrees)
            - depth (km)
            - mag Magnitude
            - mrr Mrr moment tensor component
            - mtt Mtt moment tensor component
            - mpp Mpp moment tensor component
            - mrt Mrt moment tensor component
            - mrp Mrp moment tensor component
            - mtp Mtp moment tensor component
    """
    conn = sqlite3.connect(datafile)
    cursor = conn.cursor()
    dataframe = pd.read_sql("SELECT * FROM earthquake", conn)

    # get the data source information
    dataframe["source"] = ""
    usources = dataframe["sourceid"].unique()
    for source in usources:
        cursor.execute("SELECT source FROM source WHERE id=%i" % source)
        tsource = cursor.fetchone()[0]
        rowidx = dataframe["sourceid"] == source
        dataframe.loc[rowidx, "source"] = tsource

    dataframe.drop("sourceid", axis=1, inplace=True)

    conn.close()
    return dataframe



#!/usr/bin/env python

import pandas as pd
from datetime import datetime
import tempfile
import sqlite3
import os.path

from strec.database import stash_dataframe, fetch_dataframe


def test_stash():
    d = {'time': [pd.Timestamp(datetime.utcnow()), pd.Timestamp(datetime.utcnow())],
         'lat': [34.123, 17.123],
         'lon': [-118.123, 120.123],
         'depth': [51.4, 12.7],
         'mag': [7.5, 6.4],
         'mrr': [1.2e26, 1.4],
         'mpp': [2.3e26, 1.3],
         'mtt': [3.4e26, 2.5],
         'mrt': [4.5e26, 6.5],
         'mrp': [5.6e26, 4.3],
         'mtp': [6.7e26, 2.7]}
    df = pd.DataFrame(d)

    d1 = {'time': [pd.Timestamp(datetime.utcnow()), pd.Timestamp(datetime.utcnow())],
          'lat': [34.125, 17.123],
          'lon': [-118.323, 120.123],
          'depth': [51.2, 12.7],
          'mag': [7.5, 6.4],
          'mrr': [1.2e26, 1.4],
          'mpp': [2.3e26, 1.3],
          'mtt': [3.4e26, 2.5],
          'mrt': [4.5e26, 6.5],
          'mrp': [5.6e26, 4.3],
          'mtp': [6.7e26, 2.7]}
    df1 = pd.DataFrame(d1)

    d2 = {'time': [pd.Timestamp(datetime.utcnow()), pd.Timestamp(datetime.utcnow())],
          'lat': [34.127, 17.128],
          'lon': [-118.323, 120.123],
          'depth': [51.2, 12.7],
          'mag': [7.5, 6.4],
          'mrr': [1.2e26, 1.4],
          'mpp': [2.3e26, 1.3],
          'mtt': [3.4e26, 2.5],
          'mrt': [4.5e26, 6.5],
          'mrp': [5.6e26, 4.3],
          'mtp': [6.7e26, 2.7]}
    df2 = pd.DataFrame(d2)

    dfile = None
    try:
        f, dfile = tempfile.mkstemp()
        os.close(f)
        stash_dataframe(df, dfile, 'gcmt', create_db=True)

        stash_dataframe(df1, dfile, 'usmww', create_db=False)

        stash_dataframe(df2, dfile, 'usmww', create_db=False)

        conn = sqlite3.connect(dfile)

        cursor = conn.cursor()

        cursor.execute('SELECT mrr FROM earthquake')
        rows = cursor.fetchall()
        assert rows[0][0] == 1.2e26

        df2 = fetch_dataframe(dfile)
        assert df2['mtt'][0] == df['mtt'][0]
        cursor.close()
        conn.close()
    except AssertionError as ae:
        raise AssertionError(str(ae))
    except Exception as e:
        pass
    finally:
        if dfile is not None:
            os.remove(dfile)


if __name__ == '__main__':
    test_stash()



#!/usr/bin/env python

# stdlib imports
import sys
from enum import Enum
from typing import Tuple

# third party imports
import typer

# local imports
from strec.calc import select_regions
from strec.utils import render_row

app = typer.Typer(rich_markup_mode="rich")


class FormatType(str, Enum):
    pretty = "pretty"
    csv = "csv"
    excel = "excel"
    json = "json"


@app.command()
def main(
    eqinfo: Tuple[float, float, float] = typer.Option(
        (None, None, None),
        "--eqinfo",
        "-e",
        help=("Calculate region based on lat, lon, depth of earthquake"),
    ),
    event_id: str = typer.Option(
        None,
        "--event-id",
        "-d",
        help=("Calculate region based on ComCat event id (i.e., us6000iasi)."),
    ),
    input_file: str = typer.Option(
        None,
        "--input-file",
        "-i",
        help="Calculate region for events specified in input file.",
    ),
    hypo_columns: Tuple[str, str, str] = typer.Option(
        (None, None, None),
        "--hypo_columns",
        "-c",
        help=(
            (
                "When used with -i/--input-file, specify the columns "
                "that should be used for Latitude, Longitude, Depth."
            )
        ),
    ),
    id_column: str = typer.Option(
        None,
        "--id-column",
        help=(
            (
                "When used with -i/--input-file, specify the column "
                "that should be used for ComCatID."
            )
        ),
    ),
    format: FormatType = typer.Option(
        FormatType.pretty,
        "--format",
        "-f",
        help="Specify output format for results.",
    ),
    output_file: str = typer.Option(
        None,
        "--output-file",
        "-o",
        help="Specify output filename for results (see --format).",
    ),
    moment_info: Tuple[float, float, float, float] = typer.Option(
        (None, None, None, None),
        "--moment-info",
        "-m",
        help="Specify strike, dip, rake and magnitude for single earthquake.",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Print out more verbose output.",
    ),
):
    """Determine various seismo-tectonic parameters for given input coordinates.

    The output will consist of the following data fields:

     - TectonicRegion : (Active, Stable, HotSpot, Subduction)
     - FocalMechanism : (RS [Reverse],SS [Strike-Slip], NM [Normal], ALL)
     - TensorType : composite
     - TensorSource : (composite, one of many moment tensor sources like GCMT)
     - KaganAngle : Angle between fault plane and slab orientation.
     - CompositeVariability :  A measure of the uncertainty in the composite
                               moment tensor.
     - NComposite : Number of events used to create composite moment tensor.
     - DistanceToStable : Distance in km from the nearest stable polygon.
     - DistanceToActive : Distance in km from the nearest active polygon.
     - DistanceToSubduction : Distance in km from the nearest subduction
                              polygon.
     - DistanceToHotSpot : Distance in km from the nearest hot spot polygon.
     - Oceanic : Boolean indicating whether we are in an oceanic region,
                 false indicates Continental.
     - DistanceToOceanic : Distance in km to nearest oceanic polygon.
     - DistanceToContinental : Distance in km to nearest continental polygon.
     - SlabModelRegion : Subduction region.
     - SlabModelDepth : Depth of slab2.0 grid at earthquake location.
     - SlabModelDepthUncertainty : Uncertainty of depth to slab.
     - SlabModelDip : Dip of slab at earthquake location.
     - SlabModelStrike : Strike of slab at earthquake location.
     - SlabModelMaximumDepth : Maximum depth of seismogenic zone.
     - NGACategory: Subduction category, one of ["interface", "slab", "crustal",
                    "outer rise", "undetermined", "crustal uncertain",
                    "interface uncertain", "slab uncertain", "outer rise uncertain"]

    where:
        - Active regions are boundaries between two plates that are colliding,
          sliding past each other, or pulling apart.
        - Stable regions are regions where those things are not happening.
        - Subduction regions are those that are above the Slab2.0 grids.
        - HotSpot regions are those that sit above mantle plumes
        - ALL FocalMechanisms that are some combination of the other types


    Input files can be CSV or Excel format (regcalc will
    attempt to determine format automatically) and MUST include columns
    beginning with (case insensitive):

        Lat - Numeric latitude of input earthquake.
        Lon - Numeric longitude of input earthquake.
        Depth - Numeric depth of earthquake (km).

    Input files may optionally contain a column (also case insensitive) called:
    ComCatID - Valid ANSS ComCat ID
    Any other columns present in the input will be copied to the output.

    The output format must be one of the following:

        pretty - A human readable pretty-printed representation of the output.
        json - A JSON representation of the output
        csv - A CSV output, one earthquake per line.
        excel - Microsoft Excel format, one earthquake per line.

    Note that the -c/--columns option will override these defaults.
    """
    # error out on any invalid combinations of options
    if format in ["pretty", "json"] and output_file is not None:
        print(f"Output file is invalid when specified with format of {format}")
        sys.exit(1)

    if format in ["excel"] and output_file is None:
        print("Output format 'excel' is invalid unless an output file is specified")
        sys.exit(1)

    has_eq = int(eqinfo[0] is not None)
    has_id = int(event_id is not None)
    has_file = int(input_file is not None)
    if sum([has_eq, has_id, has_file]) > 1:
        print(
            "Options --eqinfo, --event-id, --input-file are "
            "mutually exclusive. Please choose one."
        )
        sys.exit(1)

    if (hypo_columns[0] is not None or id_column is not None) and input_file is None:
        print(
            "Options --hypo-columns and --id-column only work "
            "when input_file is specified."
        )
        sys.exit(1)

    if moment_info[0] is not None and input_file is not None:
        print("You can only specify moment information for a single earthquake.")
        sys.exit(1)

    regions_frame = select_regions(
        input_file,
        eqinfo,
        moment_info,
        event_id,
        hypo_columns,
        id_column,
        verbose,
    )
    if output_file is None:
        for idx, row in regions_frame.iterrows():
            render_row(row, format, hypo_columns)

    if output_file:
        if format == "csv":
            regions_frame.to_csv(output_file, index=False)
        else:
            regions_frame.to_excel(output_file, index=False)


if __name__ == "__main__":
    app()



#!/usr/bin/env python
# stdlib imports
import pathlib

# third party imports
import numpy as np

# local imports
from strec.slab import GridSlab, SlabCollection
from strec.utils import get_config


def test_grid_slab():
    LAT, LON = -16.342671, 167.541593
    datadir = pathlib.Path(__file__).parent / "data"
    depthgrid = datadir / "van_slab2_dep_02.23.18.grd"
    dipgrid = datadir / "van_slab2_dip_02.23.18.grd"
    strgrid = datadir / "van_slab2_str_02.23.18.grd"
    uncgrid = datadir / "van_slab2_unc_02.23.18.grd"
    grid = GridSlab(depthgrid, dipgrid, strgrid, uncgrid)
    is_inside = grid.contains(LAT, LON)
    assert is_inside

    slabinfo = grid.getSlabInfo(LAT, LON)
    if not len(slabinfo):
        raise AssertionError("Slab results are empty!")
    cmp_dict = {
        "region": "van",
        "strike": 340.6477,
        "dip": 31.101204,
        "depth": 33.961307525634766,
        "maximum_interface_depth": 49,
        "depth_uncertainty": 18.315662,
    }
    for key, value in cmp_dict.items():
        value2 = slabinfo[key]
        print(f"Comparing {key} cmp {value} and actual {value2}")
        if isinstance(value, float):
            np.testing.assert_almost_equal(value, value2, decimal=4)
        else:
            assert value == value2
    print("First test successful")

    collection = SlabCollection(datadir)
    depth = 0.0
    slabinfo = collection.getSlabInfo(LAT, LON, depth)
    if not len(slabinfo):
        raise AssertionError("Slab results are empty!")
    print("Testing against slab grid...")
    for key, value in slabinfo.items():
        assert key in cmp_dict
        if isinstance(value, str):
            assert value == cmp_dict[key]
        else:
            np.testing.assert_almost_equal(value, cmp_dict[key], decimal=1)
    print("Second test successful.")


if __name__ == "__main__":
    test_grid_slab()



#!/usr/bin/env python

# third party imports
import numpy as np

# local imports
from strec.subtype import SubductionSelector, get_focal_mechanism


def reindex(results):
    results = results.reindex(
        index=[
            "TectonicRegion",
            "FocalMechanism",
            "TensorType",
            "TensorSource",
            "KaganAngle",
            "CompositeVariability",
            "NComposite",
            "DistanceToStable",
            "DistanceToActive",
            "DistanceToSubduction",
            "DistanceToHotSpot",
            "Oceanic",
            "DistanceToOceanic",
            "DistanceToContinental",
            "SlabModelRegion",
            "SlabModelDepth",
            "SlabModelDepthUncertainty",
            "SlabModelDip",
            "SlabModelStrike",
        ]
    )
    return results


def cmp_dicts(dict1, dict2):
    missing1 = set(dict1.keys()) - set(dict2.keys())
    missing2 = set(dict2.keys()) - set(dict1.keys())
    if len(missing1) or len(missing2):
        return False, str(missing1) + str(missing2)
    # keys are identical, so look at values
    mismatched_keys = []
    for key, value in dict1.items():
        if isinstance(value, float) and np.isnan(value) and np.isnan(dict2[key]):
            continue
        if value != dict2[key]:
            mismatched_keys.append(key)
    if len(mismatched_keys):
        return False, str(mismatched_keys)
    return True, ""


def test_get_focal_mechanism():
    izmit_ss = {
        "T": {"azimuth": 41, "plunge": 18},
        "N": {"azimuth": 237, "plunge": 72},
        "P": {"azimuth": 133, "plunge": 5},
        "NP1": {"strike": 178, "dip": 74, "rake": 9},
        "NP2": {"strike": 86, "dip": 81, "rake": 164},
    }

    focal_ss = get_focal_mechanism(izmit_ss)
    assert focal_ss == "SS"

    mexico_nm = {
        "T": {"azimuth": 49, "plunge": 28},
        "N": {"azimuth": 316, "plunge": 6},
        "P": {"azimuth": 215, "plunge": 61},
        "NP1": {"strike": 155, "dip": 18, "rake": -70},
        "NP2": {"strike": 315, "dip": 73, "rake": -96},
    }
    focal_nm = get_focal_mechanism(mexico_nm)
    assert focal_nm == "NM"

    northridge_rs = {
        "T": {"azimuth": 107, "plunge": 81},
        "N": {"azimuth": 293, "plunge": 9},
        "P": {"azimuth": 202, "plunge": 1},
        "NP1": {"strike": 283, "dip": 45, "rake": 77},
        "NP2": {"strike": 122, "dip": 47, "rake": 103},
    }
    focal_rs = get_focal_mechanism(northridge_rs)
    assert focal_rs == "RS"

    bogus_all = {
        "T": {"azimuth": 151, "plunge": 35.26},
        "N": {"azimuth": 4, "plunge": 35.26},
        "P": {"azimuth": 267, "plunge": 35.26},
        "NP1": {"strike": 326, "dip": 39, "rake": 45},
        "NP2": {"strike": 198, "dip": 64, "rake": 119},
    }
    focal_all = get_focal_mechanism(bogus_all)
    assert focal_all == "ALL"


def test_get_online_tensor():
    eventid_with_tensor = "official20110311054624120_30"
    eventid_without_tensor = "us2000ati0"
    eventid_with_local_tensor = "nc72852946"
    selector = SubductionSelector()
    lat1, lon1, depth1, tensor1 = selector.getOnlineTensor(eventid_with_tensor)
    lat2, lon2, depth2, tensor2 = selector.getOnlineTensor(eventid_without_tensor)
    lat3, lon3, depth3, tensor3 = selector.getOnlineTensor(eventid_with_local_tensor)
    tensor1_cmp = {
        "T": {"value": 5.6e22, "plunge": 53.0, "azimuth": 296.0},
        "mtt": -1.509e21,
        "mpp": -1.523e22,
        "mrp": 4.837e22,
        "N": {"value": 1.2e20, "plunge": 1.0, "azimuth": 204.0},
        "mrr": 1.674e22,
        "mtp": -5.286e21,
        "P": {"value": -5.6e22, "plunge": 36.0, "azimuth": 113.0},
        "NP2": {"dip": 81.0, "rake": 91.0, "strike": 24.0},
        "mrt": 2.243e22,
        "source": "duputel_201103110546a",
        "type": "Mww",
        "NP1": {"dip": 8.0, "rake": 78.0, "strike": 192.0},
    }
    tensor3_cmp = {
        "T": {"value": 8163000000000000.0, "plunge": 8.605, "azimuth": 77.715},
        "mtt": -9703000000000000.0,
        "mpp": 7235000000000000.0,
        "mrp": -685500000000000.0,
        "N": {"value": 2383000000000000.0, "plunge": 80.784, "azimuth": 278.857},
        "mrr": 2468000000000000.0,
        "mtp": -3748000000000000.0,
        "P": {"value": -1.053e16, "plunge": 3.274, "azimuth": 168.211},
        "NP2": {
            "dip": 81.586203170542277,
            "rake": 3.7979038011027644,
            "strike": 213.24771227177959,
        },
        "mrt": 903300000000000.0,
        "source": "nc_72852946",
        "type": "Mw",
        "NP1": {
            "dip": 86.243031559360034,
            "rake": 171.5679508376947,
            "strike": 122.69120043887619,
        },
    }

    res, msg = cmp_dicts(tensor1, tensor1_cmp)
    res, msg = cmp_dicts(tensor3, tensor3_cmp)
    assert tensor2 is None


def test_subtype():
    selector = SubductionSelector()

    # sumatra
    lat = 3.295
    lon = 95.982
    depth = 30.0
    eventid = "official20041226005853450_30"
    results1 = selector.getSubductionType(lat, lon, depth, eventid)
    assert results1["FocalMechanism"] == "RS"

    # northridge
    lat = 34.213
    lon = -118.537
    depth = 18.2
    eventid = "ci3144585"
    results1 = selector.getSubductionType(lat, lon, depth, eventid)
    assert results1["FocalMechanism"] == "RS"

    # landers
    lat = 34.200
    lon = -116.437
    depth = -0.1
    eventid = "ci3031111"
    results1 = selector.getSubductionType(lat, lon, depth, eventid)
    assert results1["FocalMechanism"] == "SS"

    # Made up event that should  have a composite moment tensor
    lat = 2.321
    lon = 128.132
    depth = 21.7
    results1 = selector.getSubductionType(lat, lon, depth)
    assert results1["TensorType"] == "composite"
    assert results1["FocalMechanism"] == "RS"

    # Made up event that should not have a composite moment tensor
    lat = -27.725
    lon = -147.832
    depth = 25
    results1 = selector.getSubductionType(lat, lon, depth)
    assert results1["FocalMechanism"] == "ALL"

    # Pass in an event with tensor parameters
    # chile 2010
    lat = -36.122
    lon = -72.898
    depth = 22.9
    tensor_params = {
        "source": "duputel",
        "type": "Mww",
        "NP1": {"strike": 178, "dip": 77, "rake": 86},
        "NP2": {"strike": 17, "dip": 14, "rake": 108},
        "T": {"value": 2.236e22, "plunge": 58, "azimuth": 82},
        "N": {"value": 0.040e22, "plunge": 4, "azimuth": 179},
        "P": {"value": -2.276e22, "plunge": 32, "azimuth": 272},
    }
    results1 = selector.getSubductionType(
        lat, lon, depth, eventid=None, tensor_params=tensor_params
    )
    assert results1["FocalMechanism"] == "RS"

    # event location inside grid but not horizontally on subducting slab
    lat = 13.58
    lon = -92.92
    depth = 20.0
    results1 = selector.getSubductionType(lat, lon, depth)
    assert results1["SlabModelRegion"] == "Central America"

    # event at a VERY high latitude to see if grid stuff blows up
    lat = 89.1
    lon = 49.1
    depth = 20.0
    results1 = selector.getSubductionType(lat, lon, depth)
    assert results1["TectonicRegion"] == "Stable"
    assert results1["FocalMechanism"] == "ALL"


def test_get_subduction_by_id():
    selector = SubductionSelector()
    # Tohoku, should have an online moment tensor
    eventid = "official20110311054624120_30"
    results = selector.getSubductionTypeByID(eventid)
    assert results["FocalMechanism"] == "RS"
    assert results["SlabModelRegion"] == "Kamchatka/Kurils/Japan"


def block_test_multiple_slabs():
    lat = 6.0
    lon = 125.0
    depth = 160
    selector = SubductionSelector()
    results = selector.getSubductionType(lat, lon, depth)


if __name__ == "__main__":
    test_get_focal_mechanism()
    test_subtype()
    test_get_subduction_by_id()
    # test_multiple_slabs()
    test_get_online_tensor()

    #



# stdlib imports
import configparser
import json
import pathlib
import re

# third party imports
import numpy as np
import pandas as pd
import pyproj
from shapely.geometry import LineString, shape
from shapely.ops import transform

STRECINI = "config.ini"
MOMENT_DB = "moment_tensors.db"

CONSTANTS = {
    "minradial_disthist": 0.01,
    "maxradial_disthist": 1.0,
    "minradial_distcomp": 0.5,
    "maxradial_distcomp": 1.0,
    "step_distcomp": 0.1,
    "depth_rangecomp": 10,
    "minno_comp": 3,
    "default_szdip": 17,
    "dstrike_interf": 30,
    "ddip_interf": 30,
    "dlambda": 60,
    "ddepth_interf": 20,
    "ddepth_intra": 10,
}


def get_longest_axis():
    root = pathlib.Path(__file__).parent / "data"
    jsonfiles = root.glob("**/*.geojson")
    longest_axis = 0
    for jsonfile in jsonfiles:
        with open(jsonfile, "r") as input:
            jdict = json.load(input)
        for record in jdict["features"]:
            polygon = shape(record["geometry"])
            clon, clat = polygon.centroid.xy
            pstr = f"+proj=aeqd +lon_0={clon[0]:.6f} +lat_0={clat[0]:.6f} +ellps=WGS84"
            out_crs = pyproj.CRS(pstr)
            in_crs = pyproj.CRS("EPSG:4326")
            pfunction = pyproj.Transformer.from_crs(
                in_crs, out_crs, always_xy=True
            ).transform
            ppolygon = transform(pfunction, polygon)
            # get the minimum bounding rectangle and zip coordinates into a list
            # of point-tuples
            mbr_points = list(
                zip(*ppolygon.minimum_rotated_rectangle.exterior.coords.xy)
            )

            # calculate the length of each side of the minimum bounding rectangle
            mbr_lengths = [
                LineString((mbr_points[i], mbr_points[i + 1])).length
                for i in range(len(mbr_points) - 1)
            ]

            # get major/minor axis measurements
            major_axis = max(mbr_lengths) / 1000
            if major_axis > longest_axis:
                longest_axis = major_axis

    return longest_axis


def get_config_file_name():
    config_file = pathlib.Path().home() / ".strec" / STRECINI
    return config_file


def create_config(datafolder):
    datafolder = pathlib.Path(datafolder)
    # create or re-create config file
    config_file = get_config_file_name()
    config = configparser.ConfigParser()
    longest_axis = get_longest_axis()
    config["DATA"] = {
        "folder": datafolder,
        "slabfolder": datafolder / "slabs",
        "dbfile": datafolder / MOMENT_DB,
        "longest_axis": longest_axis,
    }
    config["CONSTANTS"] = CONSTANTS
    config_folder = config_file.parent
    if not config_folder.exists():
        config_folder.mkdir(parents=True)
    with open(config_file, "wt") as cfile:
        config.write(cfile)
    return config


def get_config():
    """Get configuration information as a dictionary.

    'folder' should always be set to point to library data path.
    'dbfile' should be set to point to library data path unless specified in
    ~/.strec/strec.ini
    'slabfolder' should be set to point to library data path unless specified in
    ~/.strec/strec.ini

    Returns:
        config (dict): Dictionary containing fields:
            - CONSTANTS Dictionary containing constants for the application.
            - DATA Dictionary containing 'folder', 'slabfolder', and 'dbfile'.
    """
    # first look in the default path for a config file
    config_file = get_config_file_name()
    if config_file.is_file():
        config = configparser.ConfigParser()
        config.read(config_file)
        if "DATA" not in config:
            raise KeyError("STREC config file is missing the [DATA] section.")
    else:
        raise FileNotFoundError(f"Config file not found at {config_file}.")
    return config


def read_input_file(input_file, hypo_columns, id_column):
    """Read a CSV/Excel input file, return a DataFrame

    This function will deal with any moment tensor component columns it finds,
    (mrr,mtt,etc) and convert any that are integers into floats.

    Args:
        input_file (str): Path to CSV/Excel file containing lat,lon,depth columns
            and optionally moment tensor columns
            ('mrr','mtt','mpp','mrt','mrp','mtp').
    Returns:
        DataFrame: Pandas dataframe containing contents of input file.
    Raises:
        ValueError: When input file is neither a CSV nor an Excel file.
    """
    df = None
    try:
        df = pd.read_csv(input_file)
    except Exception:
        try:
            df = pd.read_excel(input_file)
        except Exception:
            raise ValueError("%s is neither a CSV nor Excel file." % input_file)

    idcol = None  # this is not required
    latcol = None
    loncol = None
    depcol = None
    if hypo_columns[0] is not None:
        if not set(hypo_columns).issubset(set(df.columns)):
            raise KeyError(f"Missing input hypo columns: {str(hypo_columns)}")
        latcol, loncol, depcol = hypo_columns
    if id_column is not None:
        if id_column not in df.columns:
            raise KeyError(f"Missing input comcat ID column: {id_column}")
        idcol = id_column

    # if the column names were not set by the user, try to find columns
    # that indicate comcatid, latitude, longitude and depth.
    latidx = df.columns.to_series().str.contains("^lat", case=False)
    if latidx.any():
        latcol = df.columns[latidx].array[0]
    lonidx = df.columns.to_series().str.contains("^lon", case=False)
    if lonidx.any():
        loncol = df.columns[lonidx].array[0]
    depidx = df.columns.to_series().str.contains("^depth", case=False)
    if depidx.any():
        depcol = df.columns[depidx].array[0]

    # the id column is not mandatory
    ididx = df.columns.to_series().str.contains("^eventid", case=False)
    if ididx.any():
        idcol = df.columns[ididx].array[0]

    if latcol is None or loncol is None or depcol is None:
        raise KeyError(
            (
                "Input table is missing identifiable columns for latitude, "
                "longitude, and depth."
            )
        )

    # convert any long integer moment component columns to floating point, because
    # otherwise pandas will complain later when re-writing row to a dataframe.  Doesn't
    # make sense, but seems necessary.
    for column in df:
        comps = ["mrr", "mtt", "mpp", "mrt", "mrp", "mtp"]
        for comp in comps:
            if column.lower().find(comp) > -1:
                col = df[column].apply(lambda x: convert_float(x))
                df[column] = col

    return (df, idcol, latcol, loncol, depcol)


def convert_float(val):
    try:
        return float(val)
    except ValueError:
        return np.nan


def render_row(row, format, hypo_columns):
    """Render a Series containing regselect output to the screen.

    Args:
        row (Series): Pandas series object.
        format (str): One of 'pretty','json','csv'.
        lat (float): Earthquake hypocentral latitude.
        lon (float): Earthquake hypocentral longitude.
        depth (float): Earthquake hypocentral depth.
    """
    if hypo_columns[0] is not None:
        lat, lon, depth = (
            row[hypo_columns[0]],
            row[hypo_columns[1]],
            row[hypo_columns[2]],
        )
    else:
        lat = row.filter(regex="^[l,L]at").values[0]
        lon = row.filter(regex="^[l,L]on").values[0]
        depth = row.filter(regex="^[d,D]epth").values[0]
    if format == "pretty":
        print("For event located at %.4f,%.4f,%.1f:" % (lat, lon, depth))
        for idx, value in row.items():
            if re.match("^lat|^lon|^depth", idx, re.IGNORECASE):
                continue
            print("\t%s : %s" % (idx, str(value)))
        print()
    elif format == "json":
        print(row.to_json())
    elif format == "csv":
        values = [str(v) for v in row.values]
        print(",".join(values))
